from sklearn.metrics import accuracy_score, average_precision_score, balanced_accuracy_score, f1_score, precision_score, recall_score, roc_auc_score, confusion_matrix
from _hx_model_evaluation_tools import HxFeatureImportancesandCorrelations
from _hx_model_evaluation_tools.hx_model_evaluation_graphics import HxModelEvaluationGraphics
from constants_and_tools import ConstantsAndTools
from typing import Dict, Any, List
import plotly.express as px
import pandas as pd
import numpy as np


class BinaryClassifierMetricsCalculations:
    def __init__(self,
                 probs_result_df,
                 model_save_path: str,
                 x_test_df: pd.DataFrame,
                 model,
                 importances: list,
                 target_col_name: str,
                 model_name="a",
                 bins_dict=None):
        """
        Este metodo se va a encargar de obtener las métricas de las target, su estratificación, llamar al hx_shap_tools si es preciso y devolver la informacion.
        Claves: Si se pasan las listas real_value_list, pos_prob_list, o pred_val_list con info, se usará su contenido. Sino, poe defecto se usarán los nombres que
        se han definido en las propiedades del constructor

        :param probs_result_df: Diccionario que contiene el y_test real, la probabilidad positiva predicha y la probabilidad negativa predicha
        :param model_save_path:
        :param x_test_df:
        :param model:
        :param importances:
        :param target_col_name:
        :param model_name:
        :param bins_dict:
        """

        # ----------------------------------------------------------------------------------------------------
        # -- 1: Instancio toolkits que voy a nacesitar
        # ----------------------------------------------------------------------------------------------------

        # ---- 1.1: Instancio constants
        self.CT: ConstantsAndTools = ConstantsAndTools()

        # ---- 1.2: Los gráficos de métricas
        self.GT: HxModelEvaluationGraphics = HxModelEvaluationGraphics()

        # ----------------------------------------------------------------------------------------------------
        # -- 2: Almaceno parámetros en propiedades
        # ----------------------------------------------------------------------------------------------------

        # ---- 2.1: Copio el df que contiene el valor real de test, la probabilidad positiva y la negativa
        self.probs_result_df: pd.DataFrame = probs_result_df.copy()

        # ---- 2.2: Obtengo el nombre de la columna target
        self.target_col_name: str = target_col_name

        # ---- 2.3: Creo el dataframe de metricas para estratificacion
        self.stratify_df: pd.DataFrame = pd.DataFrame(columns=["confidence"])

        # ---- 2.4: Creamos la columna pred_value_{self.target_col_name} que contiene las prediciones del modelo redondeadas
        self.probs_result_df[f"pred_value_{self.target_col_name}"] = np.round(self.probs_result_df[f'positive_proba_{self.target_col_name}'])

        # ---- 2.5: Almaceno el path donde se van a guardar los informes (si procede)
        self.model_save_path: str = model_save_path

        # ---- 2.6: Almaceno el df de test sin la target en una propiedad
        self.X_test_df: pd.DataFrame = x_test_df

        # ---- 2.7: Almaceno la instancia del modelo en una propiedad
        self.model = model

        # ---- 2.8: Almaceno la lista de pesos del modelo
        self.importances: List = importances

        # ---- 2.9: Almaceno el nombre del modelo
        self.model_name: str = model_name

        # ---- 2.10: Almaceno el diccionario de bins que se va a usar para estratificar
        self.bins_dict: dict = bins_dict

    def run(self):
        """
        Metodo que realiza la evaluacion completa de un modelo
        1: Obtiene los pesos
        2: Obtiene las métricas
        3: Obtiene la estratificacion
        :return:
        """
        # ----------------------------------------------------------------------------------------------------
        # -- 1: Realizo la evaluacion de pesos y correlaciones entre variables
        # ----------------------------------------------------------------------------------------------------

        # ---- 1.1: En caso de que existan pesos en la propiedad (Lista con los pesos) ejecutamis
        if self.importances is not None and self.importances[0] is not None:

            try:
                HxFeatureImportancesandCorrelations(self.X_test_df, self.importances[0]).run()
            except TypeError:
                try:
                    HxFeatureImportancesandCorrelations(self.X_test_df, self.importances).run()
                except TypeError:
                    pass
                except ValueError:
                    pass
            except ValueError:
                pass

        # ----------------------------------------------------------------------------------------------------
        # -- 2: Ejecutamos evaluate classification para realizar todas las tareas de evaluacion y retornamos
        # ----------------------------------------------------------------------------------------------------

        # ---- 2.1: Llamo al metodo de evaluar clasificacion y obtengo las metricas
        metrics_dict: dict = self.evaluate_clasification()

        # ---- 2.2: Retorno las metricas (no lo hago directo porque me facilita el debug si es necesario)
        return metrics_dict

    def evaluate_clasification(self):
        """
        Metodo para obtener un diccionario de métricas de cada columna target
        :return:
        """
        # ----------------------------------------------------------------------------------------------------
        # -- 1: Defino diccionario maestro y obtengo columna de aciertos
        # ----------------------------------------------------------------------------------------------------

        # ---- 1.1: Creo el diccionario de metricas
        all_metrics_dict: dict = {}

        # ---- 1.2: Cruzo la columna de la probabilidad positiva con el valor real (y_test) para obtener los aciertos del modelo
        self.probs_result_df[f"aciertos_{self.target_col_name}"] = np.where(
            (self.probs_result_df[f"real_value_{self.target_col_name}"] == 1) & (self.probs_result_df[f"positive_proba_{self.target_col_name}"] >= 0.5) |
            (self.probs_result_df[f"real_value_{self.target_col_name}"] == 0) & (self.probs_result_df[f"positive_proba_{self.target_col_name}"] < 0.5),
            True,  # Acierto si ambas condiciones son ciertas
            False  # Fallo si no coinciden
        )

        # ----------------------------------------------------------------------------------------------------
        # -- 2: Obtengo a mano la matriz de confusion (necesito mantenerlos en flotante para el stratify)
        # ----------------------------------------------------------------------------------------------------

        # ---- 2.1: Obtengo un dataframe con los Aciertos: las filas donde el acierto es True
        df_true: pd.DataFrame = self.probs_result_df[self.probs_result_df[f"aciertos_{self.target_col_name}"]]

        # ---- 2.2: Obtengo los VP: El modelo predice si y es si (VERDADEROS POSITIVOS)
        true_positive = self.probs_result_df[(self.probs_result_df[f"real_value_{self.target_col_name}"] == 1.0) &
                                             (self.probs_result_df[f"aciertos_{self.target_col_name}"])].shape[0]

        # ---- 2.3: Obtengo los VN: El modelo predice no y es no (VERDADEROS NEGATIVOS)
        true_negative = self.probs_result_df[(self.probs_result_df[f"real_value_{self.target_col_name}"] == 0.0) &
                                             (self.probs_result_df[f"aciertos_{self.target_col_name}"])].shape[0]

        # ---- 2.4: Obtengo los FN: El modelo predice no y en realidad es si (FALSOS NEGATIVOS)
        false_negative = self.probs_result_df[(self.probs_result_df[f"real_value_{self.target_col_name}"] == 1.0) &
                                              (~self.probs_result_df[f"aciertos_{self.target_col_name}"])].shape[0]

        # ---- 2.5: Obtengo los FP: El modelo predice si y en realidad es no (FALSOS POSITIVOS)
        false_positive = self.probs_result_df[(self.probs_result_df[f"real_value_{self.target_col_name}"] == 0.0) &
                                              (~self.probs_result_df[f"aciertos_{self.target_col_name}"])].shape[0]

        # ----------------------------------------------------------------------------------------------------
        # -- 3: Calculo y muestro las métricas
        # ----------------------------------------------------------------------------------------------------

        # ---- 3.1: Pinto la entrada en consola
        self.CT.IT.sub_intro_print(f"ClassifierMetricsCalculation 1: Métricas {self.model_name} {self.target_col_name}")

        # ---- 3.2: Accuracy
        accuracy = round(df_true.shape[0] / self.probs_result_df.shape[0] * 100, 3)
        self.CT.IT.info_print(f"Accuracy:      {accuracy} % -------  Formula: ({true_positive} + {true_negative})  / ({true_positive} + {true_negative} + "
                              f"{false_positive} + {false_negative}) * 100",
                              "light_cyan")

        # ---- 3.3: Recall (SENSIBILIDAD)
        sensibility = round(true_positive / (true_positive + false_negative) * 100, 3)
        self.CT.IT.info_print(f"Recall:        {sensibility} % -------  Formula: {true_positive} / ({true_positive} + {false_negative}) * 100",
                              "light_cyan")

        # ---- 3.4: Specificity: (ESPECIFICIDAD)
        specificity = round(true_negative / (true_negative + false_positive) * 100, 3)
        self.CT.IT.info_print(f"Specificity:   {specificity} % -------  Formula: {true_negative} / ({true_negative} + {false_positive}) * 100",
                              "light_cyan")

        # ---- 3.5: F1 (F1)
        f1 = round(((accuracy * sensibility) / (accuracy + sensibility)) * 2, 3)
        self.CT.IT.info_print(f"F1:            {f1} % -------  Formula: ({accuracy * sensibility:.2f} / {accuracy + sensibility:.2f}) * 2",
                              "light_cyan")

        # ----------------------------------------------------------------------------------------------------
        # -- 4: Calculo y muestro la estratificacion
        # ----------------------------------------------------------------------------------------------------

        # ---- 4.1: Creamos la nueva columna utilizando pd.cut(), esto depende de lo que le pasamos con parametro en el bins_dict
        self.probs_result_df[f'stratification_{self.target_col_name}'] = pd.cut(self.probs_result_df[f"positive_proba_{self.target_col_name}"],
                                                                                bins=self.bins_dict["bins"],
                                                                                labels=self.bins_dict["labels"],
                                                                                right=True)

        # ---- 4.2: Creo el stratify df que se va a ir agregando al self.stratify df
        stratify_df_base_cols: list = ["confidence", "total_predicho", "aciertos", "fallos", "TP_total_TP", "FP_total_FP", "TN_total_TN", "FN_total_FN"]
        self.stratify_df: pd.DataFrame = pd.DataFrame(columns=[f'{z}_{self.target_col_name}' if z != "confidence" else f"{z}" for z in stratify_df_base_cols])

        # ---- 4.3: Relleno  iterativamente los bins de estratificacion en base al self.bins_dict
        for bn in self.bins_dict["labels"][::-1]:
            if bn in self.probs_result_df[f"stratification_{self.target_col_name}"].unique().tolist():
                res_shape: int = self.probs_result_df.shape[0]
                bn_df = self.probs_result_df[self.probs_result_df[f"stratification_{self.target_col_name}"] == bn]
                bn_df_aciertos = bn_df[bn_df[f"aciertos_{self.target_col_name}"]]
                bn_df_fallos = bn_df[(~bn_df[f"aciertos_{self.target_col_name}"])]
                tp_df = bn_df_aciertos[(bn_df_aciertos[f'real_value_{self.target_col_name}'] == 1.0) | (bn_df_aciertos[f'real_value_{self.target_col_name}'] == 1)]
                fp_df = bn_df_fallos[(bn_df_fallos[f'real_value_{self.target_col_name}'] == 0.0) | (bn_df_fallos[f'real_value_{self.target_col_name}'] == 0)]
                tn_df = bn_df_aciertos[(bn_df_aciertos[f'real_value_{self.target_col_name}'] == 0.0) | (bn_df_aciertos[f'real_value_{self.target_col_name}'] == 0)]
                fn_df = bn_df_fallos[(bn_df_fallos[f'real_value_{self.target_col_name}'] == 1.0) | (bn_df_fallos[f'real_value_{self.target_col_name}'] == 1)]

                self.stratify_df.loc[len(self.stratify_df)] = {
                    f"confidence": bn,
                    f"total_predicho_{self.target_col_name}": f"{round(bn_df.shape[0] / res_shape * 100, 2)}%  ({bn_df.shape[0]}/{res_shape})" if res_shape != 0 else pd.NA,
                    f"aciertos_{self.target_col_name}": f"{round(bn_df_aciertos.shape[0] / bn_df.shape[0] * 100, 2)}%  ({bn_df_aciertos.shape[0]}/{bn_df.shape[0]})" if bn_df.shape[0] != 0 else pd.NA,
                    f"fallos_{self.target_col_name}": f"{round(bn_df_fallos.shape[0] / bn_df.shape[0] * 100, 2)}%  ({bn_df_fallos.shape[0]}/{bn_df.shape[0]})" if bn_df.shape[0] != 0 else pd.NA,
                    f"TP_total_TP_{self.target_col_name}": f"{round(tp_df.shape[0] / true_positive * 100, 2)}%  ({tp_df.shape[0]}/{true_positive})" if true_positive != 0 else pd.NA,
                    f"FP_total_FP_{self.target_col_name}": f"{round(fp_df.shape[0] / false_positive * 100, 2)}%  ({fp_df.shape[0]}/{false_positive})" if false_positive != 0 else pd.NA,
                    f"TN_total_TN_{self.target_col_name}": f"{round(tn_df.shape[0] / true_negative * 100, 2)}%  ({tn_df.shape[0]}/{true_negative})" if true_negative != 0 else pd.NA,
                    f"FN_total_FN_{self.target_col_name}": f"{round(fn_df.shape[0] / false_negative * 100, 2)}%  ({fn_df.shape[0]}/{false_negative})" if false_negative != 0 else pd.NA,

                }

        # ----------------------------------------------------------------------------------------------------
        # -- 5: Creo gráficos, pinto y retorno
        # ----------------------------------------------------------------------------------------------------

        # ---- 5.1: Creo el gráfico de estratificacion en barras
        self.plot_stratify_barplot(self.probs_result_df, f"{self.model_save_path}/stratify_barplot_{self.target_col_name}.html",
                                   f"stratification_{self.target_col_name}", f"aciertos_{self.target_col_name}")

        # ---- 5.2: Creo el diccionario con todas las metricas que voy a retornar
        all_metrics_dict[f"{self.target_col_name}"] = {"metrics": {"accuracy": accuracy,
                                                     "recall": sensibility,
                                                     "specifity": specificity,
                                                     "f1": f1,
                                                     "ALL": {"TP": true_positive, "FP": false_positive,
                                                             "TN": true_negative, "FN": false_negative},
                                                     }
                                         }

        # ---- 5.3: Pinto la entrada y las metricas de estratificacion
        self.CT.IT.sub_intro_print(f"ClassifierMetricsCalculation 2: Tabla de estratificacion {self.model_name}")
        self.CT.IT.print_tabulate_df(self.stratify_df, row_print=100)

        # ---- 5.4: Almaceno el gráfico de la estratificacion en formato tabla
        self.GT.plot_dataframe_html(self.stratify_df, save_path=f"{self.model_save_path}/stratify_df.html", save=True)

        # ---- 5.5: Retorno el diccionario completo de metricas
        return all_metrics_dict

    # <editor-fold desc="Metodos secundarios usados en evaluateClasification    --------------------------------------------------------------------------------------------------">


    @staticmethod
    def plot_stratify_barplot(df: pd.DataFrame, path_and_name: str, stratification_colname: str = "stratification", accurate_colname: str = "aciertos", show: bool = False):
        # Calcular acumulados
        df['count'] = 1
        df['acumulado'] = df.groupby(stratification_colname, observed=False)['count'].cumsum()

        # Calcular porcentajes de aciertos y fallos por rango
        df['porcentaje_aciertos'] = df.groupby(stratification_colname, observed=False)[accurate_colname].transform(lambda x: x.mean() * 100)
        df['porcentaje_fallos'] = 100 - df['porcentaje_aciertos']

        # Crear el gráfico con Plotly
        fig = px.bar(df.groupby([stratification_colname, accurate_colname], observed=False).agg({'count': 'sum'}).reset_index(),
                     x=stratification_colname, y='count', color=accurate_colname,
                     title='Cantidad de filas por rango y acierto',
                     labels={stratification_colname: 'Rangos', accurate_colname: 'Acierto/Fallo'},
                     color_discrete_map={1: '#3182bd', 0: '#de2d26'},  # Colores elegantes
                     barmode='stack',  # Apilar los colores dentro de cada rango
                     category_orders={accurate_colname: [1, 0]})  # Invertir orden de las categorías

        # Personalizar diseño del gráfico
        fig.update_layout(title={'text': 'Cantidad de filas por rango y acierto',
                                 'x': 0.5, 'y': 0.95,  # Alineación central del título
                                 'font': {'size': 20, 'color': 'black'},  # Tamaño y color de la fuente del título
                                 'pad': {'b': 30}},  # Espaciado inferior del título
                          xaxis_title='Rangos',
                          yaxis_title='Cantidad',
                          legend_title='Acierto/Fallo',
                          font=dict(size=12),  # Tamaño de la fuente de los ejes
                          hoverlabel=dict(font_size=14, font_family="Arial"),  # Tamaño y fuente del hover
                          hovermode="closest")  # Hover solo sobre el eje x

        # Calcular total por rango (inside update_traces)
        total_por_rango = df.groupby(stratification_colname, observed=False)['count'].sum()
        df = df.merge(total_por_rango.rename('total_por_rango'), left_on=stratification_colname, right_index=True)

        # Actualizar el tooltip con el total fijo
        fig.update_traces(hovertemplate='<b>Rango</b>: %{x}<br>' +
                                        '<b>Cantidad</b>: %{y} de ' +
                                        f'{df.shape[0]} totales<br>')

        fig.write_html(f"{path_and_name}")
        if show:
            # Mostrar el gráfico
            fig.show()

    # </editor-fold>


class EvaluateBinaryClassifier:
    def __init__(self, data_dict: dict, model_name: str, selected_metric: str | None = None):

        # --------------------------------------------------------------------------------------------
        # -- 1: Instancio clases de toolkits y almaceno en propiedades
        # --------------------------------------------------------------------------------------------

        # ---- 1.1: Instancio constants
        self.CT: ConstantsAndTools = ConstantsAndTools()

        # ---- 1.2: Almaceno el nombre del modelo
        self.model_name: str = model_name

        # ---- 1.3: Almaceno la metrica selecconada
        self.selected_metric: str = selected_metric

        # ---- 1.4: Almaceno los valores reales del target
        self.y_train = data_dict["y_train"]
        self.y_eval = data_dict["y_eval"]
        self.y_test = data_dict["y_test"]

        # ---- 1.5: Almaceno los valores predichos del target
        self.y_pred_train = data_dict["y_pred_train"]
        self.y_pred_eval = data_dict["y_pred_eval"]
        self.y_pred_test = data_dict["y_pred_test"]

        # ---- 1.6: Almaceno las columnas target en la propiedad (en esta versión, el multitarget no está habilitado, pero dejo la estructura)
        self.target_col_name: list = data_dict["target_col_name"]

        # ---- 1.7: Inicializo en None el df que va a contener las métricas
        self.metrics_df = None

    def calculate_and_print_metrics(self):

        # --------------------------------------------------------------------------------------------
        # -- 1: Pinto y calculo las metricas
        # --------------------------------------------------------------------------------------------

        # ---- 1.1: Pinto la entrada
        self.CT.IT.sub_intro_print("EvaluateBinaryClassifier: Calculando metricas de entrenamiento...")

        # ---- 1.2: Itero por las target_col, calculo metricas y agrego al df. (Actualmente solo hay una target)

        # -- Accuracy
        accuracy_score_train = accuracy_score(self.y_train, self.y_pred_train)
        accuracy_score_eval = accuracy_score(self.y_eval, self.y_pred_eval) if self.y_eval is not None else None
        accuracy_score_test = accuracy_score(self.y_test, self.y_pred_test)

        # -- Puntuacion de precision promedio
        avg_precision_score_train = average_precision_score(self.y_train, self.y_pred_train)
        avg_precision_score_eval = average_precision_score(self.y_eval, self.y_pred_eval) if self.y_eval is not None else None
        avg_precision_score_test = average_precision_score(self.y_test, self.y_pred_test)

        # -- Puntuacion de precision balanceada
        balanced_acc_score_train = balanced_accuracy_score(self.y_train, self.y_pred_train)
        balanced_acc_score_eval = balanced_accuracy_score(self.y_eval, self.y_pred_eval) if self.y_eval is not None else None
        balanced_acc_score_test = balanced_accuracy_score(self.y_test, self.y_pred_test)

        # -- Valor f1 (equilibrio entre accuracy y recall)
        confusion_matrix_train = confusion_matrix(self.y_train, self.y_pred_train)
        confusion_matrix_eval = confusion_matrix(self.y_eval, self.y_pred_eval) if self.y_eval is not None else None
        confusion_matrix_test = confusion_matrix(self.y_test, self.y_pred_test)
        specificity_train = confusion_matrix_train[0, 0] / (confusion_matrix_train[0, 0] + confusion_matrix_train[0, 1])
        specificity_eval = confusion_matrix_eval[0, 0] / (confusion_matrix_eval[0, 0] + confusion_matrix_eval[0, 1]) if self.y_eval is not None else None
        specificity_test = confusion_matrix_test[0, 0] / (confusion_matrix_test[0, 0] + confusion_matrix_test[0, 1])

        # -- Valor f1 (equilibrio entre accuracy y recall)
        f1_score_train = f1_score(self.y_train, self.y_pred_train)
        f1_score_eval = f1_score(self.y_eval, self.y_pred_eval) if self.y_eval is not None else None
        f1_score_test = f1_score(self.y_test, self.y_pred_test)

        # -- Numero de verdaderos positivos / total de positivos predichos (Zero division es para poner un valor cuando no se predicen muestras en una clase)
        precision_score_train = precision_score(self.y_train, self.y_pred_train, zero_division=1)
        precision_score_eval = precision_score(self.y_eval, self.y_pred_eval, zero_division=1) if self.y_eval is not None else None
        precision_score_test = precision_score(self.y_test, self.y_pred_test, zero_division=1)

        # -- Numero de verdaderos positivos / total de positivos reales
        recall_score_train = recall_score(self.y_train, self.y_pred_train)
        recall_score_eval = recall_score(self.y_eval, self.y_pred_eval) if self.y_eval is not None else None
        recall_score_test = recall_score(self.y_test, self.y_pred_test)

        # -- Area bajo la curva roc
        roc_auc_score_train = roc_auc_score(self.y_train, self.y_pred_train)
        roc_auc_score_eval = roc_auc_score(self.y_eval, self.y_pred_eval) if self.y_eval is not None else None
        roc_auc_score_test = roc_auc_score(self.y_test, self.y_pred_test)

        # -- Creo el df de metricas del individuo concreto
        hx_round = lambda x: round(x, 4) if x is not None else pd.NA
        self.metrics_df: pd.DataFrame = pd.DataFrame(columns=("metric", f"train_{self.target_col_name}", f"eval_{self.target_col_name}", f"test_{self.target_col_name}"))
        self.metrics_df.loc[len(self.metrics_df)] = ["accuracy", f"{hx_round(accuracy_score_train)}", hx_round(accuracy_score_eval), hx_round(accuracy_score_test)]
        self.metrics_df.loc[len(self.metrics_df)] = ["recall", f"{hx_round(recall_score_train)}", hx_round(recall_score_eval), hx_round(recall_score_test)]
        self.metrics_df.loc[len(self.metrics_df)] = ["specificity", f"{hx_round(specificity_train)}", hx_round(specificity_eval), hx_round(specificity_test)]
        self.metrics_df.loc[len(self.metrics_df)] = ["f1", f"{hx_round(f1_score_train)}", hx_round(f1_score_eval), hx_round(f1_score_test)]
        self.metrics_df.loc[len(self.metrics_df)] = ["roc_auc", f"{hx_round(roc_auc_score_train)}", hx_round(roc_auc_score_eval), hx_round(roc_auc_score_test)]
        self.metrics_df.loc[len(self.metrics_df)] = ["average_precision", f"{hx_round(avg_precision_score_train)}", hx_round(avg_precision_score_eval), hx_round(avg_precision_score_test)]
        self.metrics_df.loc[len(self.metrics_df)] = ["balanced_acc", f"{hx_round(balanced_acc_score_train)}", hx_round(balanced_acc_score_eval), hx_round(balanced_acc_score_test)]
        self.metrics_df.loc[len(self.metrics_df)] = ["precision", f"{hx_round(precision_score_train)}", hx_round(precision_score_eval), hx_round(precision_score_test)]

        # --------------------------------------------------------------------------------------------
        # -- 2: Proceso y transformo el metrics_df para dejarlo en el formato adecuado
        # --------------------------------------------------------------------------------------------

        # ---- 2.1: Eliminao columnas que contengan únicamente valores NaN
        self.metrics_df = self.metrics_df.dropna(axis=1, how="all")

        # ---- 2.2: Convertir a float64 todas las columnas excepto 'metric'
        self.metrics_df.loc[:, self.metrics_df.columns != 'metric'] = self.metrics_df.loc[:, self.metrics_df.columns != 'metric'].astype('float64')

        # --------------------------------------------------------------------------------------------
        # -- 3: Pinto el df y generoel diccionario de métricas
        # --------------------------------------------------------------------------------------------

        # ---- 3.1: Pinto el df
        self.CT.IT.print_tabulate_df(self.metrics_df)

        # ---- 3.2: Creo el diccionario de metricas para graficarlo finalmente en el genethic con todos los individuos
        gen_metric_dict: dict = {
            "acc_train": self.metrics_df[self.metrics_df["metric"] == "accuracy"][f"train_{self.target_col_name}"].iloc[0],
            "acc_test": self.metrics_df[self.metrics_df["metric"] == "accuracy"][f"test_{self.target_col_name}"].iloc[0],
            "recall_train": self.metrics_df[self.metrics_df["metric"] == "recall"][f"train_{self.target_col_name}"].iloc[0],
            "recall_test": self.metrics_df[self.metrics_df["metric"] == "recall"][f"test_{self.target_col_name}"].iloc[0],
            "spec_train": self.metrics_df[self.metrics_df["metric"] == "specificity"][f"train_{self.target_col_name}"].iloc[0],
            "spec_test": self.metrics_df[self.metrics_df["metric"] == "specificity"][f"test_{self.target_col_name}"].iloc[0],
            "f1_train": self.metrics_df[self.metrics_df["metric"] == "f1"][f"train_{self.target_col_name}"].iloc[0],
            "f1_test": self.metrics_df[self.metrics_df["metric"] == "f1"][f"test_{self.target_col_name}"].iloc[0],
            "roc_train": self.metrics_df[self.metrics_df["metric"] == "roc_auc"][f"train_{self.target_col_name}"].iloc[0],
            "roc_test": self.metrics_df[self.metrics_df["metric"] == "roc_auc"][f"test_{self.target_col_name}"].iloc[0],
            "bal_acc_train": self.metrics_df[self.metrics_df["metric"] == "balanced_acc"][f"train_{self.target_col_name}"].iloc[0],
            "bal_acc_test": self.metrics_df[self.metrics_df["metric"] == "balanced_acc"][f"test_{self.target_col_name}"].iloc[0],
            "precision_train": self.metrics_df[self.metrics_df["metric"] == "precision"][f"train_{self.target_col_name}"].iloc[0],
            "precision_test": self.metrics_df[self.metrics_df["metric"] == "precision"][f"test_{self.target_col_name}"].iloc[0],
        }

        del self.metrics_df

        # ---- 3.3: Devuelvo el diccionario con la metrica seleccionada y las metricas totales
        result: Dict[str, Any] = {}

        if self.selected_metric is not None:
            match self.selected_metric:
                case "accuracy":
                    result: Dict[str, Any] = {"metric_train": gen_metric_dict["acc_train"], "metric_eval": accuracy_score_eval, "metric_test": gen_metric_dict["acc_test"],
                            "all_metrics": gen_metric_dict, "selected_metric": self.selected_metric}
                case "specificity":
                    result: Dict[str, Any] = {"metric_train": gen_metric_dict["spec_train"], "metric_eval": specificity_eval, "metric_test": gen_metric_dict["spec_test"],
                            "all_metrics": gen_metric_dict, "selected_metric": self.selected_metric}
                case "balanced_acc":
                    result: Dict[str, Any] = {"metric_train": gen_metric_dict["bal_acc_train"], "metric_eval": balanced_acc_score_eval, "metric_test": gen_metric_dict["bal_acc_test"],
                            "all_metrics": gen_metric_dict, "selected_metric": self.selected_metric}
                case "f1":
                    result: Dict[str, Any] = {"metric_train": gen_metric_dict["f1_train"], "metric_eval": f1_score_eval, "metric_test": gen_metric_dict["f1_test"],
                            "all_metrics": gen_metric_dict, "selected_metric": self.selected_metric}
                case "precision":
                    result: Dict[str, Any] = {"metric_train": gen_metric_dict["precision_train"], "metric_eval": precision_score_eval, "metric_test": gen_metric_dict["precision_test"],
                            "all_metrics": gen_metric_dict, "selected_metric": self.selected_metric}
                case "recall":
                    result: Dict[str, Any] = {"metric_train": gen_metric_dict["recall_train"], "metric_eval": recall_score_eval, "metric_test": gen_metric_dict["recall_test"],
                            "all_metrics": gen_metric_dict, "selected_metric": self.selected_metric}
                case "roc_auc":
                    result: Dict[str, Any] = {"metric_train": gen_metric_dict["roc_train"], "metric_eval": roc_auc_score_eval, "metric_test": gen_metric_dict["roc_test"],
                            "all_metrics": gen_metric_dict, "selected_metric": self.selected_metric}

                case _:
                    raise ValueError("La metrica seleccionada no está contemplada")

            # ---- 3.4: Retorno el resultado
            return result

        return None


from sklearn.metrics import accuracy_score, balanced_accuracy_score, f1_score, precision_score, recall_score, roc_auc_score, confusion_matrix
from sklearn.preprocessing import label_binarize
from _hx_model_evaluation_tools import HxFeatureImportancesandCorrelations
from _hx_model_evaluation_tools.hx_model_evaluation_graphics import HxModelEvaluationGraphics
from constants_and_tools import ConstantsAndTools
from typing import Dict, Any, List
import plotly.express as px
import pandas as pd
import numpy as np


class MulticlassClassifierMetricsCalculations:
    def __init__(self,
                 probs_result_df,
                 model_save_path: str,
                 x_test_df: pd.DataFrame,
                 model,
                 importances: list,
                 target_col_name: str,
                 model_name="a",
                 bins_dict=None,
                 n_classes: int = 2):
        """
        Este metodo se va a encargar de obtener las métricas de las target, su estratificación, llamar al hx_shap_tools si es preciso y devolver la informacion.
        Adaptado para clasificación multiclase.

        :param probs_result_df: Diccionario que contiene el y_test real, las probabilidades predichas para cada clase
        :param model_save_path: Ruta donde guardar los modelos
        :param x_test_df: DataFrame con features de test
        :param model: Modelo entrenado
        :param importances: Lista de importancias de features
        :param target_col_name: Nombre de la columna target
        :param model_name: Nombre del modelo
        :param bins_dict: Diccionario con bins para estratificación
        :param n_classes: Número de clases (por defecto 2 para binaria)
        """
        # ----------------------------------------------------------------------------------------------------
        # -- 1: Instancio toolkits que voy a necesitar
        # ----------------------------------------------------------------------------------------------------

        # ---- 1.1: Instancio constants
        self.CT: ConstantsAndTools = ConstantsAndTools()

        # ---- 1.2: Los gráficos de métricas
        self.GT: HxModelEvaluationGraphics = HxModelEvaluationGraphics()

        # ----------------------------------------------------------------------------------------------------
        # -- 2: Almaceno parámetros en propiedades
        # ----------------------------------------------------------------------------------------------------

        # ---- 2.1: Copio el df que contiene el valor real de test y las probabilidades
        self.probs_result_df: pd.DataFrame = probs_result_df.copy()

        # ---- 2.2: Obtengo el nombre de la columna target
        self.target_col_name: str = target_col_name

        # ---- 2.3: Número de clases
        self.n_classes: int = n_classes

        # ---- 2.4: Creo el dataframe de metricas para estratificacion
        self.stratify_df: pd.DataFrame = pd.DataFrame(columns=["confidence"])

        # ---- 2.5: Creamos la columna pred_value_{self.target_col_name} que contiene las predicciones del modelo (clase con mayor probabilidad)
        proba_columns = [col for col in self.probs_result_df.columns if f'proba_class_' in col]

        # Para multiclase: seleccionar la clase con mayor probabilidad
        proba_array = self.probs_result_df[proba_columns].values
        self.probs_result_df[f"pred_value_{self.target_col_name}"] = np.argmax(proba_array, axis=1)

        # ---- 2.6: Almaceno el path donde se van a guardar los informes (si procede)
        self.model_save_path: str = model_save_path

        # ---- 2.7: Almaceno el df de test sin la target en una propiedad
        self.X_test_df: pd.DataFrame = x_test_df

        # ---- 2.8: Almaceno la instancia del modelo en una propiedad
        self.model = model

        # ---- 2.9: Almaceno la lista de pesos del modelo
        self.importances: List = importances

        # ---- 2.10: Almaceno el nombre del modelo
        self.model_name: str = model_name

        # ---- 2.11: Almaceno el diccionario de bins que se va a usar para estratificar
        self.bins_dict: dict = bins_dict

    def run(self):
        """
        Metodo que realiza la evaluacion completa de un modelo
        1: Obtiene los pesos
        2: Obtiene las métricas
        3: Obtiene la estratificacion
        :return: Diccionario con métricas
        """
        # ----------------------------------------------------------------------------------------------------
        # -- 1: Realizo la evaluacion de pesos y correlaciones entre variables
        # ----------------------------------------------------------------------------------------------------

        # ---- 1.1: En caso de que existan pesos en la propiedad (Lista con los pesos) ejecutamos
        if self.importances is not None and self.importances[0] is not None:
            try:
                HxFeatureImportancesandCorrelations(self.X_test_df, self.importances[0]).run()
            except (TypeError, ValueError):
                try:
                    HxFeatureImportancesandCorrelations(self.X_test_df, self.importances).run()
                except (TypeError, ValueError):
                    pass

        # ----------------------------------------------------------------------------------------------------
        # -- 2: Ejecutamos evaluate classification para realizar todas las tareas de evaluacion y retornamos
        # ----------------------------------------------------------------------------------------------------

        # ---- 2.1: Llamo al metodo de evaluar clasificacion y obtengo las metricas
        metrics_dict: dict = self.evaluate_clasification()

        # ---- 2.2: Retorno las metricas
        return metrics_dict

    def evaluate_clasification(self):
        """
        Metodo para obtener un diccionario de métricas de cada columna target
        Adaptado para multiclase
        :return: Diccionario con métricas
        """
        # ----------------------------------------------------------------------------------------------------
        # -- 1: Defino diccionario maestro y obtengo columna de aciertos
        # ----------------------------------------------------------------------------------------------------

        # ---- 1.1: Creo el diccionario de metricas
        all_metrics_dict: dict = {}

        # ---- 1.2: Cruzo la columna del valor real con el valor predicho para obtener los aciertos del modelo
        self.probs_result_df[f"aciertos_{self.target_col_name}"] = (
                self.probs_result_df[f"real_value_{self.target_col_name}"] ==
                self.probs_result_df[f"pred_value_{self.target_col_name}"]
        )

        # ----------------------------------------------------------------------------------------------------
        # -- 2: Obtengo la matriz de confusion para multiclase
        # ----------------------------------------------------------------------------------------------------

        # ---- 2.1: Obtengo los valores reales y predichos
        y_true = self.probs_result_df[f"real_value_{self.target_col_name}"].values
        y_pred = self.probs_result_df[f"pred_value_{self.target_col_name}"].values

        # ---- 2.2: Calculo la matriz de confusión
        cm = confusion_matrix(y_true, y_pred, labels=range(self.n_classes))

        # ---- 2.3: Calculo métricas por clase
        class_metrics = {}
        for i in range(self.n_classes):
            # Verdaderos positivos para la clase i
            tp = cm[i, i]

            # Falsos positivos para la clase i (predicciones de clase i que no son i)
            fp = cm[:, i].sum() - tp

            # Falsos negativos para la clase i (reales de clase i predichos como otras clases)
            fn = cm[i, :].sum() - tp

            # Verdaderos negativos para la clase i (taquello que no es clase i correctamente identificado)
            tn = cm.sum() - (tp + fp + fn)

            class_metrics[f"class_{i}"] = {
                "TP": int(tp),
                "FP": int(fp),
                "TN": int(tn),
                "FN": int(fn),
                "accuracy": round(float((tp + tn) / (tp + tn + fp + fn)), 2) if (tp + tn + fp + fn) > 0 else 0,
                "recall": round(float(tp / (tp + fn)), 2) if (tp + fn) > 0 else 0,
                "precision": round(float(tp / (tp + fp)), 2) if (tp + fp) > 0 else 0,
                "specificity": round(float(tn / (tn + fp)), 2) if (tn + fp) > 0 else 0
            }

        # ----------------------------------------------------------------------------------------------------
        # -- 3: Calculo métricas globales
        # ----------------------------------------------------------------------------------------------------

        # ---- 3.1: Pinto la entrada en consola
        self.CT.IT.sub_intro_print(f"ClassifierMetricsCalculation 1: Métricas {self.model_name} {self.target_col_name} - {self.n_classes} clases")

        # ---- 3.2: Accuracy global
        accuracy = round(accuracy_score(y_true, y_pred) * 100, 2)
        self.CT.IT.info_print(f"Accuracy Global:      {accuracy:.2f} %", "light_cyan")

        # ---- 3.3: Balanced Accuracy (promedio de recall por clase)
        balanced_accuracy = round(balanced_accuracy_score(y_true, y_pred) * 100, 2)
        self.CT.IT.info_print(f"Balanced Accuracy:    {balanced_accuracy:.2f} %", "light_cyan")

        # ---- 3.4: F1 Macro (promedio de F1 por clase)
        f1_macro = round(f1_score(y_true, y_pred, average='macro') * 100, 2)
        self.CT.IT.info_print(f"F1 Macro:             {f1_macro:.2f} %", "light_cyan")

        # ---- 3.5: F1 Micro (F1 global considerando todos los TP, FP, FN)
        f1_micro = round(f1_score(y_true, y_pred, average='micro') * 100, 2)
        self.CT.IT.info_print(f"F1 Micro:             {f1_micro:.2f} %", "light_cyan")

        # ---- 3.6: Precision Macro
        precision_macro = round(precision_score(y_true, y_pred, average='macro', zero_division=0) * 100, 2)
        self.CT.IT.info_print(f"Precision Macro:      {precision_macro:.2f} %", "light_cyan")

        # ---- 3.7: Recall Macro
        recall_macro = round(recall_score(y_true, y_pred, average='macro', zero_division=0) * 100, 2)
        self.CT.IT.info_print(f"Recall Macro:         {recall_macro:.2f} %", "light_cyan")

        # ----------------------------------------------------------------------------------------------------
        # -- 4: Calculo y muestro la estratificacion (usando la probabilidad de la clase predicha)
        # ----------------------------------------------------------------------------------------------------

        # ---- 4.1: Para multiclase, usamos la máxima probabilidad como confianza
        proba_columns = [col for col in self.probs_result_df.columns if f'proba_class_' in col]

        max_proba = self.probs_result_df[proba_columns].max(axis=1)
        self.probs_result_df['max_proba'] = max_proba

        # ---- 4.2: Creamos la estratificación basada en la máxima probabilidad
        self.probs_result_df[f'stratification_{self.target_col_name}'] = pd.cut(
            self.probs_result_df['max_proba'],
            bins=self.bins_dict["bins"],
            labels=self.bins_dict["labels"],
            right=False
        )

        # ---- 4.3: Creo el stratify df
        stratify_df_base_cols: list = ["confidence", "total_predicho", "aciertos", "fallos"]
        # Agregar columnas por clase
        for i in range(self.n_classes):
            stratify_df_base_cols.extend([f"class_{i}_correct", f"class_{i}_incorrect"])

        self.stratify_df: pd.DataFrame = pd.DataFrame(
            columns=[f'{z}_{self.target_col_name}' if z != "confidence" else f"{z}" for z in stratify_df_base_cols]
        )

        # ---- 4.4: Relleno iterativamente los bins de estratificacion
        total_rows = self.probs_result_df.shape[0]
        for bn in self.bins_dict["labels"][::-1]:
            if bn in self.probs_result_df[f"stratification_{self.target_col_name}"].unique().tolist():
                bn_df = self.probs_result_df[self.probs_result_df[f"stratification_{self.target_col_name}"] == bn]
                bn_df_aciertos = bn_df[bn_df[f"aciertos_{self.target_col_name}"]]
                bn_df_fallos = bn_df[~bn_df[f"aciertos_{self.target_col_name}"]]

                # Datos básicos del bin
                row_data = {
                    f"confidence": bn,
                    f"total_predicho_{self.target_col_name}": f"{round(bn_df.shape[0] / total_rows * 100, 2)}%  ({bn_df.shape[0]}/{total_rows})",
                    f"aciertos_{self.target_col_name}": f"{round(bn_df_aciertos.shape[0] / bn_df.shape[0] * 100, 2)}%  ({bn_df_aciertos.shape[0]}/{bn_df.shape[0]})" if bn_df.shape[
                                                                                                                                                                            0] != 0 else pd.NA,
                    f"fallos_{self.target_col_name}": f"{round(bn_df_fallos.shape[0] / bn_df.shape[0] * 100, 2)}%  ({bn_df_fallos.shape[0]}/{bn_df.shape[0]})" if bn_df.shape[
                                                                                                                                                                      0] != 0 else pd.NA,
                }

                # Datos por clase
                for i in range(self.n_classes):
                    class_correct = bn_df_aciertos[bn_df_aciertos[f"pred_value_{self.target_col_name}"] == i].shape[0]
                    class_incorrect = bn_df_fallos[bn_df_fallos[f"real_value_{self.target_col_name}"] == i].shape[0]

                    row_data[f"class_{i}_correct_{self.target_col_name}"] = f"{class_correct}"
                    row_data[f"class_{i}_incorrect_{self.target_col_name}"] = f"{class_incorrect}"

                self.stratify_df.loc[len(self.stratify_df)] = row_data

        # ----------------------------------------------------------------------------------------------------
        # -- 5: Creo gráficos, pinto y retorno
        # ----------------------------------------------------------------------------------------------------

        # ---- 5.1: Creo el gráfico de estratificacion en barras (adaptado para multiclase)
        self.plot_stratify_barplot_multiclass(
            self.probs_result_df,
            f"{self.model_save_path}/stratify_barplot_{self.target_col_name}.html",
            f"stratification_{self.target_col_name}",
            f"aciertos_{self.target_col_name}",
            f"pred_value_{self.target_col_name}",
            self.n_classes
        )

        # ---- 5.2: Creo el diccionario con todas las metricas
        all_metrics_dict[f"{self.target_col_name}"] = {
            "metrics": {
                "accuracy": accuracy,
                "balanced_accuracy": balanced_accuracy,
                "f1_macro": f1_macro,
                "f1_micro": f1_micro,
                "precision_macro": precision_macro,
                "recall_macro": recall_macro,
                "confusion_matrix": cm.tolist(),
                "class_metrics": class_metrics
            }
        }

        # ---- 5.3: Pinto la entrada y las metricas de estratificacion
        self.CT.IT.sub_intro_print(f"ClassifierMetricsCalculation 2: Tabla de estratificacion {self.model_name}")
        self.CT.IT.print_tabulate_df(self.stratify_df, row_print=100)

        # ---- 5.4: Almaceno el gráfico de la estratificacion en formato tabla
        self.GT.plot_dataframe_html(self.stratify_df, save_path=f"{self.model_save_path}/stratify_df.html", save=True)

        # ---- 5.5: Retorno el diccionario completo de metricas
        return all_metrics_dict

    # <editor-fold desc="Metodos secundarios usados en evaluateClasification    --------------------------------------------------------------------------------------------------">

    @staticmethod
    def plot_stratify_barplot_multiclass(df: pd.DataFrame, path_and_name: str, stratification_colname: str,
                                         accurate_colname: str, pred_class_colname: str, n_classes: int, show: bool = False):
        """
        Gráfico de barras para multiclase
        """
        # Calcular conteos por clase y estratificación
        df['count'] = 1
        results = []

        for strat in df[stratification_colname].unique():
            for class_idx in range(n_classes):
                strat_class_df = df[(df[stratification_colname] == strat) & (df[pred_class_colname] == class_idx)]
                correct = strat_class_df[accurate_colname].sum()
                incorrect = len(strat_class_df) - correct

                results.append({
                    'stratification': strat,
                    'class': f'Class {class_idx}',
                    'correct': correct,
                    'incorrect': incorrect
                })

        results_df = pd.DataFrame(results)

        # Crear gráfico apilado
        fig = px.bar(results_df, x='stratification', y=['correct', 'incorrect'],
                     color_discrete_map={'correct': '#3182bd', 'incorrect': '#de2d26'},
                     title='Distribución de aciertos por clase y rango de confianza',
                     labels={'stratification': 'Rango de Confianza', 'value': 'Cantidad', 'variable': 'Resultado'},
                     barmode='stack')

        fig.update_layout(title={'text': 'Distribución de aciertos por clase y rango de confianza',
                                 'x': 0.5, 'y': 0.95,
                                 'font': {'size': 20, 'color': 'black'},
                                 'pad': {'b': 30}})

        fig.write_html(path_and_name)
        if show:
            fig.show()

    # </editor-fold>


class EvaluateMulticlassClassifier:
    def __init__(self, data_dict: dict, model_name: str, selected_metric: str | None = None, n_classes: int = 2):

        # --------------------------------------------------------------------------------------------
        # -- 1: Instancio clases de toolkits y almaceno en propiedades
        # --------------------------------------------------------------------------------------------

        # ---- 1.1: Instancio constants
        self.CT: ConstantsAndTools = ConstantsAndTools()

        # ---- 1.2: Almaceno el nombre del modelo
        self.model_name: str = model_name

        # ---- 1.3: Almaceno la metrica seleccionada
        self.selected_metric: str = selected_metric

        # ---- 1.4: Almaceno el número de clases
        self.n_classes: int = n_classes

        # ---- 1.5: Almaceno los valores reales del target
        self.y_train = data_dict["y_train"]
        self.y_eval = data_dict["y_eval"]
        self.y_test = data_dict["y_test"]

        # ---- 1.6: Almaceno los valores predichos del target
        self.y_pred_train = data_dict["y_pred_train"]
        self.y_pred_eval = data_dict["y_pred_eval"]
        self.y_pred_test = data_dict["y_pred_test"]

        # ---- 1.7: Almaceno las columnas target
        self.target_col_name: list = data_dict["target_col_name"]

        # ---- 1.8: Inicializo en None el df que va a contener las métricas
        self.metrics_df = None

    def calculate_and_print_metrics(self):
        """
        Calcula métricas para clasificación multiclase
        :return: Diccionario con métricas
        """
        # --------------------------------------------------------------------------------------------
        # -- 1: Pinto y calculo las metricas
        # --------------------------------------------------------------------------------------------

        # ---- 1.1: Pinto la entrada
        self.CT.IT.sub_intro_print("EvaluateClassifier: Calculando metricas de entrenamiento (Multiclase)...")

        # --------------------------------------------------------------------------------------------
        # -- 2: Calculo métricas para multiclase
        # --------------------------------------------------------------------------------------------

        # ---- 2.1: Accuracy
        accuracy_score_train = accuracy_score(self.y_train, self.y_pred_train)
        accuracy_score_eval = accuracy_score(self.y_eval, self.y_pred_eval) if self.y_eval is not None else None
        accuracy_score_test = accuracy_score(self.y_test, self.y_pred_test)

        # ---- 2.2: Balanced Accuracy
        balanced_acc_score_train = balanced_accuracy_score(self.y_train, self.y_pred_train)
        balanced_acc_score_eval = balanced_accuracy_score(self.y_eval, self.y_pred_eval) if self.y_eval is not None else None
        balanced_acc_score_test = balanced_accuracy_score(self.y_test, self.y_pred_test)

        # ---- 2.3: F1 Scores
        f1_macro_train = f1_score(self.y_train, self.y_pred_train, average='macro')
        f1_macro_eval = f1_score(self.y_eval, self.y_pred_eval, average='macro') if self.y_eval is not None else None
        f1_macro_test = f1_score(self.y_test, self.y_pred_test, average='macro')

        f1_micro_train = f1_score(self.y_train, self.y_pred_train, average='micro')
        f1_micro_eval = f1_score(self.y_eval, self.y_pred_eval, average='micro') if self.y_eval is not None else None
        f1_micro_test = f1_score(self.y_test, self.y_pred_test, average='micro')

        f1_weighted_train = f1_score(self.y_train, self.y_pred_train, average='weighted')
        f1_weighted_eval = f1_score(self.y_eval, self.y_pred_eval, average='weighted') if self.y_eval is not None else None
        f1_weighted_test = f1_score(self.y_test, self.y_pred_test, average='weighted')

        # ---- 2.4: Precision Scores
        precision_macro_train = precision_score(self.y_train, self.y_pred_train, average='macro', zero_division=0)
        precision_macro_eval = precision_score(self.y_eval, self.y_pred_eval, average='macro', zero_division=0) if self.y_eval is not None else None
        precision_macro_test = precision_score(self.y_test, self.y_pred_test, average='macro', zero_division=0)

        precision_micro_train = precision_score(self.y_train, self.y_pred_train, average='micro', zero_division=0)
        precision_micro_eval = precision_score(self.y_eval, self.y_pred_eval, average='micro', zero_division=0) if self.y_eval is not None else None
        precision_micro_test = precision_score(self.y_test, self.y_pred_test, average='micro', zero_division=0)

        # ---- 2.5: Recall Scores
        recall_macro_train = recall_score(self.y_train, self.y_pred_train, average='macro', zero_division=0)
        recall_macro_eval = recall_score(self.y_eval, self.y_pred_eval, average='macro', zero_division=0) if self.y_eval is not None else None
        recall_macro_test = recall_score(self.y_test, self.y_pred_test, average='macro', zero_division=0)

        recall_micro_train = recall_score(self.y_train, self.y_pred_train, average='micro', zero_division=0)
        recall_micro_eval = recall_score(self.y_eval, self.y_pred_eval, average='micro', zero_division=0) if self.y_eval is not None else None
        recall_micro_test = recall_score(self.y_test, self.y_pred_test, average='micro', zero_division=0)

        # ---- 2.6: ROC AUC  (one-vs-rest para multiclase)

        y_train_bin = label_binarize(self.y_train, classes=range(self.n_classes))
        y_test_bin = label_binarize(self.y_test, classes=range(self.n_classes))

        roc_auc_train = roc_auc_score(y_train_bin, label_binarize(self.y_pred_train, classes=range(self.n_classes)), average='macro', multi_class='ovr')
        roc_auc_test = roc_auc_score(y_test_bin, label_binarize(self.y_pred_test, classes=range(self.n_classes)), average='macro', multi_class='ovr')
        roc_auc_eval = roc_auc_score(label_binarize(self.y_eval, classes=range(self.n_classes)),
                                     label_binarize(self.y_pred_eval, classes=range(self.n_classes)),
                                     average='macro', multi_class='ovr') if self.y_eval is not None else None

        # ---- 2.7: Matrices de confusión
        confusion_matrix_train = confusion_matrix(self.y_train, self.y_pred_train)
        confusion_matrix_eval = confusion_matrix(self.y_eval, self.y_pred_eval) if self.y_eval is not None else None
        confusion_matrix_test = confusion_matrix(self.y_test, self.y_pred_test)

        # --------------------------------------------------------------------------------------------
        # -- 3: Creo el dataframe de métricas
        # --------------------------------------------------------------------------------------------

        hx_round = lambda x: round(x, 4) if x is not None else pd.NA

        self.metrics_df: pd.DataFrame = pd.DataFrame(columns=("metric", f"train_{self.target_col_name}", f"eval_{self.target_col_name}", f"test_{self.target_col_name}"))

        # Agregar todas las métricas
        metrics_data = [
            ("accuracy", hx_round(accuracy_score_train), hx_round(accuracy_score_eval), hx_round(accuracy_score_test)),
            ("balanced_accuracy", hx_round(balanced_acc_score_train), hx_round(balanced_acc_score_eval), hx_round(balanced_acc_score_test)),
            ("f1_macro", hx_round(f1_macro_train), hx_round(f1_macro_eval), hx_round(f1_macro_test)),
            ("f1_micro", hx_round(f1_micro_train), hx_round(f1_micro_eval), hx_round(f1_micro_test)),
            ("f1_weighted", hx_round(f1_weighted_train), hx_round(f1_weighted_eval), hx_round(f1_weighted_test)),
            ("precision_macro", hx_round(precision_macro_train), hx_round(precision_macro_eval), hx_round(precision_macro_test)),
            ("precision_micro", hx_round(precision_micro_train), hx_round(precision_micro_eval), hx_round(precision_micro_test)),
            ("recall_macro", hx_round(recall_macro_train), hx_round(recall_macro_eval), hx_round(recall_macro_test)),
            ("recall_micro", hx_round(recall_micro_train), hx_round(recall_micro_eval), hx_round(recall_micro_test)),
            ("roc_auc_ovr", hx_round(roc_auc_train), hx_round(roc_auc_eval), hx_round(roc_auc_test))
        ]

        for metric_name, train_val, eval_val, test_val in metrics_data:
            self.metrics_df.loc[len(self.metrics_df)] = [metric_name, train_val, eval_val, test_val]

        # --------------------------------------------------------------------------------------------
        # -- 4: Proceso y transformo el metrics_df
        # --------------------------------------------------------------------------------------------

        # ---- 4.1: Elimino columnas que contengan únicamente valores NaN
        self.metrics_df = self.metrics_df.dropna(axis=1, how="all")

        # ---- 4.2: Convertir a float64 todas las columnas excepto 'metric'
        self.metrics_df.loc[:, self.metrics_df.columns != 'metric'] = self.metrics_df.loc[:, self.metrics_df.columns != 'metric'].astype('float64')

        # --------------------------------------------------------------------------------------------
        # -- 5: Pinto el df y genero el diccionario de métricas
        # --------------------------------------------------------------------------------------------

        # ---- 5.1: Pinto el df
        self.CT.IT.print_tabulate_df(self.metrics_df)

        # ---- 5.2: Creo el diccionario de metricas
        gen_metric_dict: dict = {}
        for metric in self.metrics_df["metric"].unique():
            gen_metric_dict[f"{metric}_train"] = self.metrics_df[self.metrics_df["metric"] == metric][f"train_{self.target_col_name}"].iloc[0]
            gen_metric_dict[f"{metric}_test"] = self.metrics_df[self.metrics_df["metric"] == metric][f"test_{self.target_col_name}"].iloc[0]
            if f"eval_{self.target_col_name}" in self.metrics_df.columns:
                gen_metric_dict[f"{metric}_eval"] = self.metrics_df[self.metrics_df["metric"] == metric][f"eval_{self.target_col_name}"].iloc[0]

        # Agregar matrices de confusión
        gen_metric_dict["confusion_matrix_train"] = confusion_matrix_train.tolist()
        gen_metric_dict["confusion_matrix_test"] = confusion_matrix_test.tolist()
        if confusion_matrix_eval is not None:
            gen_metric_dict["confusion_matrix_eval"] = confusion_matrix_eval.tolist()

        del self.metrics_df

        # ---- 5.3: Devuelvo el diccionario con la metrica seleccionada y las metricas totales
        result: Dict[str, Any] = {}

        if self.selected_metric is not None:
            metric_map = {
                "accuracy": ("accuracy", "accuracy"),
                "balanced_accuracy": ("balanced_accuracy", "balanced_accuracy"),
                "f1_macro": ("f1_macro", "f1_macro"),
                "f1_micro": ("f1_micro", "f1_micro"),
                "f1_weighted": ("f1_weighted", "f1_weighted"),
                "precision_macro": ("precision_macro", "precision_macro"),
                "recall_macro": ("recall_macro", "recall_macro"),
                "roc_auc": ("roc_auc_ovr", "roc_auc")
            }

            if self.selected_metric in metric_map:
                metric_key, display_name = metric_map[self.selected_metric]
                result = {
                    "metric_train": gen_metric_dict.get(f"{metric_key}_train"),
                    "metric_eval": gen_metric_dict.get(f"{metric_key}_eval"),
                    "metric_test": gen_metric_dict.get(f"{metric_key}_test"),
                    "all_metrics": gen_metric_dict,
                    "selected_metric": display_name
                }
            else:
                raise ValueError(f"La métrica seleccionada '{self.selected_metric}' no está contemplada para multiclase")

            return result

        return None


from hx_machine_learning_tools.regressor_ml_models.HxLightGbmRegressor import HxLightGbmRegressor
from hx_machine_learning_tools.regressor_ml_models.HxXtremeGradientBoostingRegressor import HxXtremeGradientBoostingRegressor
from hx_machine_learning_tools.regressor_ml_models.HxCatBoostRegressor import HxCatBoostRegressor


from hx_hyperparameters_builder.abstract_hyperparameter_builder import AbstractHyperparameterBuilder
from typing import Dict, Union

# <editor-fold desc="MACHINE LEARNING --------------------------------------------------------------------------------------------------------------------------------------------">

class HxLightgbmHyperparameterBuilder(AbstractHyperparameterBuilder):
    def __init__(self, custom_hyperparameters: Dict[str, Union[int, float]] | None = None):

        # -------------------------------------------------------------------------------
        # -- 1: Llamo a la clase padre con la configuracion concreta del modelo
        # -------------------------------------------------------------------------------

        # ---- 1.1: Hago el call y asigno los valores por defecto a las propiedades recogidas en la clase padre
        super().__init__(
            {'max_depth': (1, 4), 'learning_rate': (0.005, 0.2), 'n_estimators': (100, 3000), 'reg_alpha': (0, 3), 'reg_lambda': (0, 3)},
            ["learning_rate", "min_split_gain"],
            {'max_depth': (1, 15), 'learning_rate': (0.001, 0.5), 'n_estimators': (100, 8000), 'reg_alpha': (0, 9), 'reg_lambda': (0, 9)}
        )

        # -------------------------------------------------------------------------------
        # -- 2: Obtengo los custom_hyperparameters
        # -------------------------------------------------------------------------------

        # ---- 2.1: Almaceno en una propiedad los hiperparametros que ha definido el usuario
        self.custom_hyperparameters: Dict[str, Union[int, float]] | None = custom_hyperparameters

    def validate_malformation_ranges(self):
        pass

    def validate_and_get_hyperparams(self) -> Dict[str, Union[int, float]]:
        """
        Metodo que va a devolver los custom_hyperparameters si no son None y va a lanzar una excepcion en caso contrario
        :return:
        """

        # -- Planteo la excepción
        if (self.custom_hyperparameters is None or
                not isinstance(self.custom_hyperparameters, dict) or
                not all(isinstance(k, str) for k in self.custom_hyperparameters.keys()) or
                not all(isinstance(v, (int, float)) for v in self.custom_hyperparameters.values())):
            raise ValueError("custom_hyperparameters must be Dict[str, Union[int, float]]. Example: {'max_depth': 3, 'learning_rate': 0.01}")

        # -- Los retorno si están bien
        return self.custom_hyperparameters

class HxXgboostHyperparameterBuilder(AbstractHyperparameterBuilder):
    def __init__(self, custom_hyperparameters: Dict[str, Union[int, float]] | None = None):

        # -------------------------------------------------------------------------------
        # -- 1: Llamo a la clase padre con la configuracion concreta del modelo
        # -------------------------------------------------------------------------------

        # ---- 1.1: Hago el call y asigno los valores por defecto a las propiedades recogidas en la clase padre
        super().__init__(
            {'max_depth': (1, 10), 'learning_rate': (0.005, 0.2), 'n_estimators': (100, 3000), 'reg_alpha': (0, 3), 'reg_lambda': (0, 3)},
            ["learning_rate", "min_split_gain"],
            {'max_depth': (1, 15), 'learning_rate': (0.001, 0.5), 'n_estimators': (100, 8000), 'reg_alpha': (0, 9), 'reg_lambda': (0, 9)}
        )

        # -------------------------------------------------------------------------------
        # -- 2: Obtengo los custom_hyperparameters
        # -------------------------------------------------------------------------------

        # ---- 2.1: Almaceno en una propiedad los hiperparametros que ha definido el usuario
        self.custom_hyperparameters: Dict[str, Union[int, float]] | None = custom_hyperparameters

    def validate_malformation_ranges(self):
        pass

    def validate_and_get_hyperparams(self) -> Dict[str, Union[int, float]]:
        """
        Metodo que va a devolver los custom_hyperparameters si no son None y va a lanzar una excepcion en caso contrario
        :return:
        """

        # -- Planteo la excepción
        if (self.custom_hyperparameters is None or
                not isinstance(self.custom_hyperparameters, dict) or
                not all(isinstance(k, str) for k in self.custom_hyperparameters.keys()) or
                not all(isinstance(v, (int, float)) for v in self.custom_hyperparameters.values())):
            raise ValueError("custom_hyperparameters must be Dict[str, Union[int, float]]. Example: {'max_depth': 3, 'learning_rate': 0.01}")

        # -- Los retorno si están bien
        return self.custom_hyperparameters

class HxCatBoostHyperparameterBuilder(AbstractHyperparameterBuilder):
    def __init__(self, custom_hyperparameters: Dict[str, Union[int, float]] | None = None):

        # -------------------------------------------------------------------------------
        # -- 1: Llamo a la clase padre con la configuracion concreta del modelo
        # -------------------------------------------------------------------------------

        # ---- 1.1: Hago el call y asigno los valores por defecto a las propiedades recogidas en la clase padre
        super().__init__(
            {'iterations': (100, 900), 'max_depth': (1, 10), 'learning_rate': (0.005, 0.2), 'l2_leaf_reg': (0, 3), 'penalties_coefficient': (0, 3)},
            ["learning_rate"],
            {'iterations': (100, 5000), 'max_depth': (1, 15), 'learning_rate': (0.001, 0.5), 'l2_leaf_reg': (0, 9), 'penalties_coefficient': (0, 9)}
        )

        # -------------------------------------------------------------------------------
        # -- 2: Obtengo los custom_hyperparameters
        # -------------------------------------------------------------------------------

        # ---- 2.1: Almaceno en una propiedad los hiperparametros que ha definido el usuario
        self.custom_hyperparameters: Dict[str, Union[int, float]] | None = custom_hyperparameters

    def validate_malformation_ranges(self):
        pass

    def validate_and_get_hyperparams(self) -> Dict[str, Union[int, float]]:
        """
        Metodo que va a devolver los custom_hyperparameters si no son None y va a lanzar una excepcion en caso contrario
        :return:
        """

        # -- Planteo la excepción
        if (self.custom_hyperparameters is None or
                not isinstance(self.custom_hyperparameters, dict) or
                not all(isinstance(k, str) for k in self.custom_hyperparameters.keys()) or
                not all(isinstance(v, (int, float)) for v in self.custom_hyperparameters.values())):
            raise ValueError("custom_hyperparameters must be Dict[str, Union[int, float]]. Example: {'max_depth': 3, 'learning_rate': 0.01}")

        # -- Los retorno si están bien
        return self.custom_hyperparameters

class HxSvmHyperparameterBuilder(AbstractHyperparameterBuilder):
    def __init__(self, custom_hyperparameters: Dict[str, Union[int, float]] | None = None):

        # -------------------------------------------------------------------------------
        # -- 1: Llamo a la clase padre con la configuracion concreta del modelo
        # -------------------------------------------------------------------------------

        # ---- 1.1: Hago el call y asigno los valores por defecto a las propiedades recogidas en la clase padre
        super().__init__(
            {'C': (0.01, 10), 'gamma': (0.1, 0.8), 'degree': (2, 4)},
            ["C", "gamma"],
            {'C': (0.001, 10), 'gamma': (0.0, 1), 'degree': (1, 10)}
        )

        # -------------------------------------------------------------------------------
        # -- 2: Obtengo los custom_hyperparameters
        # -------------------------------------------------------------------------------

        # ---- 2.1: Almaceno en una propiedad los hiperparametros que ha definido el usuario
        self.custom_hyperparameters: Dict[str, Union[int, float]] | None = custom_hyperparameters

    def validate_malformation_ranges(self):
        pass

    def validate_and_get_hyperparams(self) -> Dict[str, Union[int, float]]:
        """
        Metodo que va a devolver los custom_hyperparameters si no son None y va a lanzar una excepcion en caso contrario
        :return:
        """

        # -- Planteo la excepción
        if (self.custom_hyperparameters is None or
                not isinstance(self.custom_hyperparameters, dict) or
                not all(isinstance(k, str) for k in self.custom_hyperparameters.keys()) or
                not all(isinstance(v, (int, float)) for v in self.custom_hyperparameters.values())):
            raise ValueError("custom_hyperparameters must be Dict[str, Union[int, float]]. Example: {'max_depth': 3, 'learning_rate': 0.01}")

        # -- Los retorno si están bien
        return self.custom_hyperparameters

# </editor-fold>

# <editor-fold desc="DEEP LEARNING -----------------------------------------------------------------------------------------------------------------------------------------------">

class HxDnnHyperparameterBuilder(AbstractHyperparameterBuilder):
    def __init__(self, custom_hyperparameters: Dict[str, Union[int, float]] | None = None):

        # -------------------------------------------------------------------------------
        # -- 1: Llamo a la clase padre con la configuracion concreta del modelo
        # -------------------------------------------------------------------------------

        # ---- 1.1: Hago el call y asigno los valores por defecto a las propiedades recogidas en la clase padre
        super().__init__(
            {'units': (248, 1024), 'dropout': (0, 0.3), 'epochs': (20, 250), 'batch_size': (16, 1024), 'learning_rate': (0.00001, 0.001),
             'regularization': (0, 0.001), 'layers_design': (0, 2.2), 'layers_number': (1, 8)},
            ["learning_rate", "dropout", "regularization"],
            {'units': (4, 4096), 'dropout': (0, 0.4), 'epochs': (10, 400), 'batch_size': (8, 2048), 'learning_rate': (0.000005, 0.5),
             'regularization': (0, 0.005), 'layers_design': (0, 4), 'layers_number': (1, 9)}
        )

        # -------------------------------------------------------------------------------
        # -- 2: Obtengo los custom_hyperparameters
        # -------------------------------------------------------------------------------

        # ---- 2.1: Almaceno en una propiedad los hiperparametros que ha definido el usuario
        self.custom_hyperparameters: Dict[str, Union[int, float]] | None = custom_hyperparameters

    def validate_malformation_ranges(self):
        pass

    def validate_and_get_hyperparams(self) -> Dict[str, Union[int, float]]:
        """
        Metodo que va a devolver los custom_hyperparameters si no son None y va a lanzar una excepcion en caso contrario
        :return:
        """

        # -- Planteo la excepción
        if (self.custom_hyperparameters is None or
                not isinstance(self.custom_hyperparameters, dict) or
                not all(isinstance(k, str) for k in self.custom_hyperparameters.keys()) or
                not all(isinstance(v, (int, float)) for v in self.custom_hyperparameters.values())):
            raise ValueError("custom_hyperparameters must be Dict[str, Union[int, float]]. Example: {'max_depth': 3, 'learning_rate': 0.01}")

        # -- Los retorno si están bien
        return self.custom_hyperparameters

class HxCnn1dHyperparameterBuilder(AbstractHyperparameterBuilder):
    def __init__(self, custom_hyperparameters: Dict[str, Union[int, float]] | None = None):

        # -------------------------------------------------------------------------------
        # -- 1: Llamo a la clase padre con la configuracion concreta del modelo
        # -------------------------------------------------------------------------------

        # ---- 1.1: Hago el call y asigno los valores por defecto a las propiedades recogidas en la clase padre
        super().__init__(
            {'filters': (8, 128), 'kernel_size': (3, 5), 'pool_size': (0, 4), 'dropout': (0, 0.2), 'epochs': (20, 50), 'batch_size': (24, 128),
             'learning_rate': (0.00001, 0.001), 'regularization': (0, 0.001), 'layers_design': (0, 2.2), 'layers_number': (0, 2), 'dense_aditional_layers': (0, 2),
             'dense_aditional_units': (32, 128)},
            ["learning_rate", "dropout", "regularization"],
            {'filters': (4, 1024), 'kernel_size': (2, 9), 'pool_size': (0, 4), 'dropout': (0, 0.3), 'epochs': (10, 120), 'batch_size': (16, 1024),
             'learning_rate': (0.000005, 0.5), 'regularization': (0, 0.005), 'layers_design': (0, 4), 'layers_number': (0, 3), 'dense_aditional_layers': (0, 3),
             'dense_aditional_units': (32, 128)}
        )

        # -------------------------------------------------------------------------------
        # -- 2: Obtengo los custom_hyperparameters
        # -------------------------------------------------------------------------------

        # ---- 2.1: Almaceno en una propiedad los hiperparametros que ha definido el usuario
        self.custom_hyperparameters: Dict[str, Union[int, float]] | None = custom_hyperparameters

    def validate_malformation_ranges(self):
        pass

    def validate_and_get_hyperparams(self) -> Dict[str, Union[int, float]]:
        """
        Metodo que va a devolver los custom_hyperparameters si no son None y va a lanzar una excepcion en caso contrario
        :return:
        """

        # -- Planteo la excepción
        if (self.custom_hyperparameters is None or
                not isinstance(self.custom_hyperparameters, dict) or
                not all(isinstance(k, str) for k in self.custom_hyperparameters.keys()) or
                not all(isinstance(v, (int, float)) for v in self.custom_hyperparameters.values())):
            raise ValueError("custom_hyperparameters must be Dict[str, Union[int, float]]. Example: {'max_depth': 3, 'learning_rate': 0.01}")

        # -- Los retorno si están bien
        return self.custom_hyperparameters

class HxLstmHyperparameterBuilder(AbstractHyperparameterBuilder):
    def __init__(self, custom_hyperparameters: Dict[str, Union[int, float]] | None = None):

        # -------------------------------------------------------------------------------
        # -- 1: Llamo a la clase padre con la configuracion concreta del modelo
        # -------------------------------------------------------------------------------

        # ---- 1.1: Hago el call y asigno los valores por defecto a las propiedades recogidas en la clase padre
        super().__init__(
            {'timesteps': (3, 7), 'filters': (8, 256), 'kernel_size': (3, 5), 'dropout': (0, 0.2), 'epochs': (20, 120), 'batch_size': (24, 512),
             'learning_rate': (0.00001, 0.001), 'regularization': (0, 0.001), 'layers_design': (0, 4), 'layers_number': (0, 3), 'dense_aditional_layers': (0, 3),
             'dense_aditional_units': (32, 512), 'lstm_aditional_layers': (0, 3), 'lstm_aditional_units': (32, 128)},
            ["learning_rate", "dropout", "regularization"],
            {'timesteps': (3, 7), 'filters': (6, 512), 'kernel_size': (2, 6), 'dropout': (0, 0.3), 'epochs': (15, 150), 'batch_size': (16, 1024),
             'learning_rate': (0.000005, 0.5), 'regularization': (0, 0.005), 'layers_design': (0, 4), 'layers_number': (0, 4), 'dense_aditional_layers': (0, 3),
             'dense_aditional_units': (24, 648), 'lstm_aditional_layers': (0, 3), 'lstm_aditional_units': (24, 164)}
        )

        # -------------------------------------------------------------------------------
        # -- 2: Obtengo los custom_hyperparameters
        # -------------------------------------------------------------------------------

        # ---- 2.1: Almaceno en una propiedad los hiperparametros que ha definido el usuario
        self.custom_hyperparameters: Dict[str, Union[int, float]] | None = custom_hyperparameters

    def validate_malformation_ranges(self):
        pass

    def validate_and_get_hyperparams(self) -> Dict[str, Union[int, float]]:
        """
        Metodo que va a devolver los custom_hyperparameters si no son None y va a lanzar una excepcion en caso contrario
        :return:
        """

        # -- Planteo la excepción
        if (self.custom_hyperparameters is None or
                not isinstance(self.custom_hyperparameters, dict) or
                not all(isinstance(k, str) for k in self.custom_hyperparameters.keys()) or
                not all(isinstance(v, (int, float)) for v in self.custom_hyperparameters.values())):
            raise ValueError("custom_hyperparameters must be Dict[str, Union[int, float]]. Example: {'max_depth': 3, 'learning_rate': 0.01}")

        # -- Los retorno si están bien
        return self.custom_hyperparameters

class HxHybridHyperparameterBuilder(AbstractHyperparameterBuilder):
    def __init__(self, custom_hyperparameters: Dict[str, Union[int, float]] | None = None):

        # -------------------------------------------------------------------------------
        # -- 1: Llamo a la clase padre con la configuracion concreta del modelo
        # -------------------------------------------------------------------------------

        # ---- 1.1: Hago el call y asigno los valores por defecto a las propiedades recogidas en la clase padre
        super().__init__(
            {'filters': (8, 256), 'kernel_size': (3, 12), 'pool_size': (0, 1), 'dropout': (0, 0.2), 'epochs': (20, 70), 'batch_size': (128, 512),
             'learning_rate': (0.00001, 0.001), 'regularization': (0, 0.001), 'layers_design': (0, 4), 'layers_number': (0, 2), 'dense_aditional_layers': (0, 2),
             'dense_aditional_units': (32, 128), 'lstm_aditional_layers': (0, 2), 'lstm_aditional_units': (32, 128)},
            ["learning_rate", "dropout", "regularization"],
            {'filters': (4, 1024), 'kernel_size': (2, 15), 'pool_size': (0, 4), 'dropout': (0, 0.3), 'epochs': (10, 120), 'batch_size': (24, 1024),
             'learning_rate': (0.000005, 0.5), 'regularization': (0, 0.005), 'layers_design': (0, 4), 'layers_number': (0, 4), 'dense_aditional_layers': (0, 3),
             'dense_aditional_units': (32, 128), 'lstm_aditional_layers': (0, 3), 'lstm_aditional_units': (24, 164)}
        )

        # -------------------------------------------------------------------------------
        # -- 2: Obtengo los custom_hyperparameters
        # -------------------------------------------------------------------------------

        # ---- 2.1: Almaceno en una propiedad los hiperparametros que ha definido el usuario
        self.custom_hyperparameters: Dict[str, Union[int, float]] | None = custom_hyperparameters

    def validate_malformation_ranges(self):
        pass

    def validate_and_get_hyperparams(self) -> Dict[str, Union[int, float]]:
        """
        Metodo que va a devolver los custom_hyperparameters si no son None y va a lanzar una excepcion en caso contrario
        :return:
        """

        # -- Planteo la excepción
        if (self.custom_hyperparameters is None or
                not isinstance(self.custom_hyperparameters, dict) or
                not all(isinstance(k, str) for k in self.custom_hyperparameters.keys()) or
                not all(isinstance(v, (int, float)) for v in self.custom_hyperparameters.values())):
            raise ValueError("custom_hyperparameters must be Dict[str, Union[int, float]]. Example: {'max_depth': 3, 'learning_rate': 0.01}")

        # -- Los retorno si están bien
        return self.custom_hyperparameters

class HxTransformerHyperparameterBuilder(AbstractHyperparameterBuilder):
    def __init__(self, custom_hyperparameters: Dict[str, Union[int, float]] | None = None):

        # -------------------------------------------------------------------------------
        # -- 1: Llamo a la clase padre con la configuracion concreta del modelo
        # -------------------------------------------------------------------------------

        # ---- 1.1: Hago el call y asigno los valores por defecto a las propiedades recogidas en la clase padre
        super().__init__(
            {'num_blocks': (1, 3), 'sequence_length': (1, 4), 'buffer_size': (100, 1500), 'd_model': (32, 256), 'num_heads': (1, 8), 'ff_dim': (1, 4),
             'embedding_size': (32, 128), 'dropout_rate': (0.1, 0.4), 'batch_size': (8, 32), 'epochs': (1, 100), 'learning_rate': (0.00001, 0.0001), 'units': (1, 50),
             'feature_dim': (1, 1), 'regularization': (0.0, 0.2), 'layers_design': (0, 4), 'layers_number': (1, 3), 'dense_aditional_layers': (0, 2),
             'dense_aditional_units': (24, 128)},
            ["learning_rate", "dropout_rate", "regularization"],
            {'num_blocks': (1, 6), 'sequence_length': (1, 5), 'buffer_size': (100, 1500), 'd_model': (32, 256), 'num_heads': (1, 8), 'ff_dim': (1, 4),
             'embedding_size': (32, 128), 'dropout_rate': (0.1, 0.4), 'batch_size': (8, 32), 'epochs': (1, 100), 'learning_rate': (0.000001, 0.001), 'units': (1, 50),
             'feature_dim': (1, 1), 'regularization': (0.0, 0.2), 'layers_design': (0, 4.25), 'layers_number': (1, 3), 'dense_aditional_layers': (0, 2),
             'dense_aditional_units': (24, 128)}
        )

        # -------------------------------------------------------------------------------
        # -- 2: Obtengo los custom_hyperparameters
        # -------------------------------------------------------------------------------

        # ---- 2.1: Almaceno en una propiedad los hiperparametros que ha definido el usuario
        self.custom_hyperparameters: Dict[str, Union[int, float]] | None = custom_hyperparameters

    def validate_malformation_ranges(self):
        pass

    def validate_and_get_hyperparams(self) -> Dict[str, Union[int, float]]:
        """
        Metodo que va a devolver los custom_hyperparameters si no son None y va a lanzar una excepcion en caso contrario
        :return:
        """

        # -- Planteo la excepción
        if (self.custom_hyperparameters is None or
                not isinstance(self.custom_hyperparameters, dict) or
                not all(isinstance(k, str) for k in self.custom_hyperparameters.keys()) or
                not all(isinstance(v, (int, float)) for v in self.custom_hyperparameters.values())):
            raise ValueError("custom_hyperparameters must be Dict[str, Union[int, float]]. Example: {'max_depth': 3, 'learning_rate': 0.01}")

        # -- Los retorno si están bien
        return self.custom_hyperparameters

# </editor-fold>


from typing import List, Set, Tuple, Dict, Optional
from constants_and_tools import ConstantsAndTools
import pandas as pd
import numpy as np


class HxFeatureImportancesandCorrelations:
    """
    Clase para identificar y eliminar variables altamente correlacionadas basándose en su importancia/peso.

    Esta clase analiza la matriz de correlación de un conjunto de datos y elimina las variables
    que están altamente correlacionadas entre sí, manteniendo únicamente la variable con mayor
    peso/importancia dentro de cada grupo correlacionado.
    """

    def __init__(self, x_test: pd.DataFrame, weights: List[float], corr_limit: float = 0.97) -> None:
        """
        Inicializa la clase con los datos y parámetros necesarios.

        Args:
            x_test (pd.DataFrame): DataFrame con las variables a analizar
            weights (List[float]): Lista con los pesos/importancia de cada variable,
                                 debe tener el mismo orden que las columnas de x_test
            corr_limit (float, optional): Umbral de correlación por encima del cual
                                        se consideran variables altamente correlacionadas.
                                        Por defecto 0.97

        Raises:
            ValueError: Si el número de pesos no coincide con el número de columnas
        """
        # Inicializar herramientas y constantes
        self.CT: ConstantsAndTools = ConstantsAndTools()

        # Validar que el número de pesos coincida con el número de columnas
        if len(weights) != len(x_test.columns):
            raise ValueError(f"El número de pesos ({len(weights)}) debe coincidir "
                             f"con el número de columnas ({len(x_test.columns)})")

        # Almacenar datos de entrada
        self.X_test: pd.DataFrame = x_test
        self.weights: List[float] = weights
        self.corr_limit: float = corr_limit

        # Calcular matriz de correlación absoluta
        self.corr_matrix: pd.DataFrame = self.X_test.corr().abs()

    def _find_high_correlation_pairs(self) -> List[Tuple[int, int]]:
        """
        Encuentra todos los pares de variables con correlación superior al límite establecido.

        Solo considera la matriz triangular superior para evitar duplicados y la diagonal.

        Returns:
            List[Tuple[int, int]]: Lista de tuplas con los índices de las columnas
                                  altamente correlacionadas
        """
        # Crear máscara para la matriz triangular superior (excluyendo diagonal)
        upper_triangle_mask = np.triu(np.ones(self.corr_matrix.shape), k=1).astype(bool)

        # Aplicar máscara a la matriz de correlación
        upper_triangle = self.corr_matrix.where(upper_triangle_mask)

        # Encontrar pares con correlación superior al límite
        high_corr_indices = np.where(upper_triangle > self.corr_limit)
        high_corr_pairs = list(zip(high_corr_indices[0], high_corr_indices[1]))

        return high_corr_pairs

    def _group_correlated_variables(self, high_corr_pairs: List[Tuple[int, int]]) -> List[Set[str]]:
        """
        Agrupa variables altamente correlacionadas en conjuntos.

        Si dos variables están correlacionadas con una tercera, todas forman un grupo.

        Args:
            high_corr_pairs (List[Tuple[int, int]]): Pares de índices de variables correlacionadas

        Returns:
            List[Set[str]]: Lista de conjuntos, cada uno conteniendo nombres de variables
                           que están altamente correlacionadas entre sí
        """
        high_corr_groups: List[Set[str]] = []

        for i, j in high_corr_pairs:
            # Convertir índices a nombres de columnas
            col1 = self.corr_matrix.columns[i]
            col2 = self.corr_matrix.columns[j]

            # Buscar si alguna de las columnas ya pertenece a un grupo existente
            found_group: Optional[Set[str]] = None
            for group in high_corr_groups:
                if col1 in group or col2 in group:
                    found_group = group
                    break

            if found_group is not None:
                # Agregar ambas columnas al grupo existente
                found_group.update([col1, col2])
            else:
                # Crear nuevo grupo con ambas columnas
                high_corr_groups.append({col1, col2})

        return high_corr_groups

    def _identify_variables_to_drop(self, high_corr_groups: List[Set[str]]) -> List[str]:
        """
        Identifica qué variables eliminar de cada grupo correlacionado.

        En cada grupo, mantiene la variable con mayor peso y marca las demás para eliminación.

        Args:
            high_corr_groups (List[Set[str]]): Grupos de variables correlacionadas

        Returns:
            List[str]: Lista de nombres de columnas a eliminar
        """
        to_drop: List[str] = []

        for group in high_corr_groups:
            # Calcular importancia/peso de cada variable en el grupo
            importances: Dict[str, float] = {
                col: self.weights[self.X_test.columns.get_loc(col)]
                for col in group
            }

            # Encontrar la variable con mayor importancia
            max_importance_col = max(importances, key=importances.get)

            # Agregar todas las demás variables del grupo a la lista de eliminación
            variables_to_remove = [col for col in group if col != max_importance_col]
            to_drop.extend(variables_to_remove)

        return to_drop

    def _print_results(self, to_drop: List[str], x_filtered: pd.DataFrame, high_corr_groups: List[Set[str]]) -> None:
        """
        Imprime los resultados del análisis de correlaciones.

        Args:
            to_drop (List[str]): Variables eliminadas
            x_filtered (pd.DataFrame): DataFrame filtrado
            high_corr_groups (List[Set[str]]): Grupos de variables correlacionadas
        """
        # Imprimir variables eliminadas
        self.CT.IT.sub_intro_print("CorrelationsAndWeights 1: Variables a eliminar por estar muy correlacionadas y tener menos peso:")
        self.CT.IT.info_print(f"{to_drop}")

        # Imprimir variables restantes
        self.CT.IT.sub_intro_print("CorrelationsAndWeights 2: Variables restantes ordenadas por peso")
        remaining_variables = list(x_filtered.columns)
        self.CT.IT.info_print(f"{remaining_variables}")

        # Imprimir grupos de variables correlacionadas
        self.CT.IT.sub_intro_print("CorrelationsAndWeights 3: Grupos de variables altamente correlacionadas entre sí")

        if len(high_corr_groups) == 0:
            self.CT.IT.info_print(f"No hay variables con alta correlacion")
        else:
            for i, group in enumerate(high_corr_groups, 1):
                self.CT.IT.info_print(f"Grupo {i}: {group}")

    def run(self) -> pd.DataFrame:
        """
        Ejecuta el proceso completo de eliminación de variables correlacionadas.

        Proceso:
        1. Encuentra pares de variables altamente correlacionadas
        2. Agrupa las variables correlacionadas
        3. Identifica cuáles eliminar basándose en sus pesos
        4. Filtra el DataFrame eliminando las variables seleccionadas
        5. Imprime resultados del análisis

        Returns:
            pd.DataFrame: DataFrame filtrado sin las variables altamente correlacionadas
                         de menor importancia
        """
        # Paso 1: Encontrar pares de variables altamente correlacionadas
        high_corr_pairs = self._find_high_correlation_pairs()

        # Paso 2: Agrupar variables correlacionadas
        high_corr_groups = self._group_correlated_variables(high_corr_pairs)

        # Paso 3: Identificar variables a eliminar
        to_drop = self._identify_variables_to_drop(high_corr_groups)

        # Paso 4: Filtrar DataFrame eliminando columnas seleccionadas
        x_filtered = self.X_test.drop(columns=to_drop)

        # Paso 5: Imprimir resultados
        self._print_results(to_drop, x_filtered, high_corr_groups)

        return x_filtered

    def get_correlation_summary(self) -> Dict[str, any]:
        """
        Obtiene un resumen del análisis de correlaciones sin modificar los datos.

        Returns:
            Dict[str, any]: Diccionario con información resumida del análisis:
                - 'high_corr_pairs': Número de pares altamente correlacionados
                - 'correlation_groups': Número de grupos de variables correlacionadas
                - 'variables_to_drop': Número de variables que serían eliminadas
                - 'remaining_variables': Número de variables que permanecerían
        """
        high_corr_pairs = self._find_high_correlation_pairs()
        high_corr_groups = self._group_correlated_variables(high_corr_pairs)
        to_drop = self._identify_variables_to_drop(high_corr_groups)

        return {
            'high_corr_pairs': len(high_corr_pairs),
            'correlation_groups': len(high_corr_groups),
            'variables_to_drop': len(to_drop),
            'remaining_variables': len(self.X_test.columns) - len(to_drop),
            'correlation_threshold': self.corr_limit
        }



from _hx_model_evaluation_tools import EvaluateRegressor, RegressorMetricsCalculations
from hx_machine_learning_tools.ml_base_model import HxMachineLearningBaseModel
from lightgbm import LGBMRegressor
from _hx_model_evaluation_tools import MlShapBinaryAndRegressorTools
from typing import Dict, Literal, List, Any, Tuple
from joblib import dump
import pandas as pd
import numpy as np
import datetime
import os


class HxLightGbmRegressor(HxMachineLearningBaseModel):

    def __init__(self,
                 data_dict: Dict[str, pd.DataFrame],
                 hiperparams: dict,
                 metric: Literal['mse', 'rmse', 'mae', 'msle', 'rmsle', 'mape'],
                 problem_type: Literal['regression', 'regression_l1', 'quantile', 'tweedie'] = 'regression_l1',
                 save_path: str | None = None,
                 bins: dict | None = None,
                 verbose: Literal['show_all', 'show_basics', 'dont_show'] = 'dont_show',
                 random_state: int = 42):
        """
        Clase abstracta para construir modelos de clasificacion de Machine Learning
        :param data_dict: Diccionario que contiene el x_train, y_train, x_test, y_test
        :param save_path:
        :param hiperparams:
        :param bins:
        """
        super().__init__(data_dict, hiperparams, "LGBM_regressor", save_path)

        # --------------------------------------------------------------------------------------------
        # -- 0: Almaceno nombre de la clase, pinto la entrada y creo la propiedad model
        # --------------------------------------------------------------------------------------------

        # ---- 0.1: Almaceno el nombre de la clase
        self.class_name: str = f"HxLightGbmRegressor"

        # ---- 0.2: Pinto la entrada
        self.CT.IT.intro_print(self.class_name)

        # ---- 0.3: Creo la propiedad model y la inicializo en None
        self.model = None

        # --------------------------------------------------------------------------------------------
        # -- 1: Asigno las propiedades abstractas
        # --------------------------------------------------------------------------------------------

        # ---- 1.1: Verbose
        self._verbose: Literal['show_all', 'show_basics', 'dont_show'] = verbose

        # ---- 1.2: Metric
        self._metric: Literal['mse', 'rmse', 'mae', 'msle', 'rmsle', 'mape'] = metric

        # ---- 1.3: Problem type
        self._problem_type: Literal['regression', 'regression_l1', 'quantile', 'tweedie'] = problem_type

        # --------------------------------------------------------------------------------------------
        # -- 2: Asigno las propiedades estandar
        # --------------------------------------------------------------------------------------------

        # ---- 2.1: Asigno el bins, por defecto en regresion es None
        self.bins = bins

        # ---- 2.2: Path donde se va a guardar la informacion del modelo
        self.model_save_path: str = self.get_save_path()

        # ---- 2.3: Random state con el que se van a inicializar los pesos, por defecto 42
        self.random_state: int = random_state

    # <editor-fold desc="Implementacion de metodos abstractos   ------------------------------------------------------------------------------------------------------------------">

    def get_save_path(self) -> str:
        """
        Metodo que hace lo siguiente:
        1: Busca dentro del path definido en la clase padre {save_path}/training_results/&Y_%m_%d/v_xxx si existe algúna carpeta llamda {self.model_name
        2: Si ya existe crea una nueva version, es decir {save_path}/training_results/&Y_%m_%d/v_xxx+1/{self.model_name y la asigna como self.save_path
        3: Si no existe, la asigna como self.save_path, es decir: {save_path}/training_results/&Y_%m_%d/v_xxx/{self.model_name
        :return: el save_path
        """

        # -- Si el path no existe, simplemente lo creo y lo asigno
        if not os.path.exists(f"{self.save_path}/{self.current_train_version}/{self.model_name}"):
            self.CT.OT.create_folder_if_not_exists(f"{self.save_path}/{self.current_train_version}/{self.model_name}")

        # -- Si el path existe, creo una nueva version y lo asigno
        else:
            self.current_train_version: str = self.new_training_version()

            self.CT.OT.create_folder_if_not_exists(f"{self.save_path}/{self.current_train_version}")
            self.CT.OT.create_folder_if_not_exists(f"{self.save_path}/{self.current_train_version}/{self.model_name}")

        # -- Actualizo las propiedades pertinentres de la clase padre
        self.day_train_versions: List[str] = self.list_versions()

        # -- Retorno el path
        return f"{self.save_path}/{self.current_train_version}/{self.model_name}"

    def fit_and_get_metrics(self, save_metrics_dict: bool) -> Dict[str, Any]:
        """
        Implementacion del metodo que entrena y devuelve el diccionario de métricas
        :param save_metrics_dict: Boleana que indica si quieres guardar el diccionario de metricas
        :return: diccionario de mettricas
        """
        # --------------------------------------------------------------------------------------------
        # -- 0: Pinto la entrada
        # --------------------------------------------------------------------------------------------
        start_date: datetime.datetime = datetime.datetime.now()
        self.CT.IT.sub_intro_print(f"{self.class_name}.fit_and_get_metrics: Inicio de entrenamiento a las {start_date}...")

        # --------------------------------------------------------------------------------------------
        # -- 1: Instancio el modelo y entreno
        # --------------------------------------------------------------------------------------------

        # ---- 1.1: Instancio el modelo y le asigno los patrámetros
        self.model: LGBMRegressor = LGBMRegressor(objective=self.problem_type, random_state=self.random_state, verbose=self.verbose, **self.hiperparams)

        # ---- 1.2: Realizo el entrenamiento
        self.model.fit(X=self.x_train, y=self.y_train.ravel())

        # ---- 1.3: Pinto el fin del entrenamiento y lo que ha tardado
        self.CT.IT.info_print(f"Entrenamiento completado a las {datetime.datetime.now()}. Ha tardado {datetime.datetime.now() - start_date}", "light_magenta")

        # --------------------------------------------------------------------------------------------
        # -- 2: Realizo las predicciones, mando a evaluar y retorno el diccionario de métricas
        # --------------------------------------------------------------------------------------------

        # ---- 2.1: Realizo las predicciones
        y_pred_train = self.model.predict(self.x_train_df)
        y_pred_test = self.model.predict(self.x_test_df)

        # ---- 2.2. Evalúo y almaceno las metricas
        metrics: Dict[str, Any] = EvaluateRegressor({"y_train": self.y_train,
                                                               "y_eval": None,
                                                               "y_test": self.y_test,
                                                               "y_pred_train": y_pred_train,
                                                               "y_pred_eval": None,
                                                               "y_pred_test": y_pred_test,
                                                               "target_col_name": self.target_col_name},
                                                              self.model_name,
                                                              self.metric).calculate_and_print_metrics()

        # ---- 2.3: Almaceno el diccionario de metricas como json en el path
        if save_metrics_dict:
            self.save_metrics_dict_json(metrics, self.model_save_path)

        # ---- 2.4. Pinto y retorno las métricas
        end_date = datetime.datetime.now()
        self.CT.IT.info_print(f"fit_and_get_metrics terminado a las {end_date}. Ha tardado {end_date - start_date}", "light_magenta")
        return metrics

    def fit_and_get_model_and_results(self) -> Tuple[Any, np.ndarray, np.ndarray, Dict[str, Any], List]:
        """
        Implementacion del metodo que entrena y devuelve (model, x_test_probs, x_train_probs, self.master_dict, model_weights)

        """
        # --------------------------------------------------------------------------------------------
        # -- 0: Pinto la entrada
        # --------------------------------------------------------------------------------------------
        start_date: datetime.datetime = datetime.datetime.now()
        self.CT.IT.sub_intro_print(f"{self.class_name}.fit_and_get_model_and_results: Inicio de entrenamiento a las {start_date}...")

        # --------------------------------------------------------------------------------------------
        # -- 1: Instancio el modelo y entreno
        # --------------------------------------------------------------------------------------------

        # ---- 1.1: Instancio el modelo y le asigno los patrámetros
        self.model: LGBMRegressor = LGBMRegressor(objective=self.problem_type, random_state=self.random_state, verbose=self.verbose, **self.hiperparams)

        # ---- 1.2: Realizo el entrenamiento
        self.model.fit(X=self.x_train, y=self.y_train.ravel())

        # ---- 1.3: Almaceno el modelo y el modulo del que proviene
        model_class_name: str = self.model.__class__.__name__
        model_module_name: str = self.model.__class__.__module__

        # ---- 1.4: Pinto el fin del entrenamiento y lo que ha tardado
        self.CT.IT.info_print(f"Entrenamiento completado a las {datetime.datetime.now()}. Ha tardado {datetime.datetime.now() - start_date}", "light_magenta")

        # --------------------------------------------------------------------------------------------
        # -- 2: Realizo las predicciones, mando a evaluar y retorno el diccionario de métricas
        # --------------------------------------------------------------------------------------------

        # ---- 2.1: Realizo las predicciones
        y_pred_train = self.model.predict(self.x_train_df)
        y_pred_test = self.model.predict(self.x_test_df)

        # ---- 2.2. Evalúo las metricas
        EvaluateRegressor(
            {
                "y_train": self.y_train,
                "y_eval": None,
                "y_test": self.y_test,
                "y_pred_train": y_pred_train,
                "y_pred_eval": None,
                "y_pred_test": y_pred_test,
                "target_col_name": self.target_col_name
            },
            self.model_name,
            self.metric).calculate_and_print_metrics()

        # --------------------------------------------------------------------------------------------
        # -- 3: Almaceno el modelo en su path con joblib.dump y obtengo los pesos
        # --------------------------------------------------------------------------------------------

        # ---- 3.1: Almacenamiento del modelo
        dump(self.model, f"{self.model_save_path}/{self.class_name}.joblib")

        # ---- 3.2: Obtencion de pesos (llamo al metodo polimorfeado self.get_weights)
        model_weights: List = self.get_weights(self.model)

        # ---- 3.3: Creacion de diccionario en el que asocio cada peso con su columna
        model_weights_dict: Dict[str, Any] = {col: weight for col, weight in zip([z for z in self.x_train_df.columns], [float(x) for x in model_weights[0]])}

        # --------------------------------------------------------------------------------------------
        # -- 4: Realizo las predicciones de probabilidad en test (valor continuo entre 0 y 1) y las almaceno
        # --------------------------------------------------------------------------------------------

        # ---- 4.1: Obtengo una matriz con las prediciones
        x_train_probs: np.ndarray[tuple[int]] = self.model.predict(self.x_train_df)
        x_test_probs: np.ndarray[tuple[int]] = self.model.predict(self.x_test_df)

        # ---- 4.2: Creo un diccionario llamado probs_result_df en el que almaceno el y_test real y los de las probs
        probs_result_dict: Dict[str, Any] = {
            f'real_value_{self.target_col_name}': self.y_test[:, 0],
            f'predict_value_{self.target_col_name}': x_test_probs
        }

        # ---- 4.3: Asigno probs_result_df a un dataframe para trabajarlo con mas facilidad
        probs_result_df: pd.DataFrame = pd.DataFrame(probs_result_dict)

        # --------------------------------------------------------------------------------------------
        # -- 5: Instancio BinaryClassifierMetricsCalculations para obtener toda la info del resultado y rellenar el self.master_dict
        # --------------------------------------------------------------------------------------------

        # ---- 5.1: Obtengo el diccionario de metricas que proporciona RegressorMetricsCalculations.run
        metrics_dict: dict = RegressorMetricsCalculations(
            probs_result_df,
            self.model_save_path,
            self.x_test_df,
            self.model,
            model_weights,
            self.target_col_name,
            self.model_name,
            self.bins
        ).run()

        # ---- 5.2: Relleno el self.master_dict
        self.master_result_dict[f"{self.model_name}"]: dict = {}
        self.master_result_dict[f"{self.model_name}"]["columns"] = [z for z in self.x_train_df.columns]
        self.master_result_dict[f"{self.model_name}"]["target"] = self.target_col_name
        self.master_result_dict[f"{self.model_name}"]["best_params"] = self.hiperparams
        self.master_result_dict[f"{self.model_name}"]["metrics"] = metrics_dict
        self.master_result_dict[f"{self.model_name}"]["model_weights"] = model_weights_dict
        self.master_result_dict[f"{self.model_name}"]["model_class_name"] = model_class_name
        self.master_result_dict[f"{self.model_name}"]["model_module_name"] = model_module_name
        self.master_result_dict[f"{self.model_name}"]["model_version"] = f"{self.today_str_daye}_{self.current_train_version}"

        # --------------------------------------------------------------------------------------------
        # -- 6: Pinto y retorno la tupla (model, x_test_probs, x_train_probs, self.master_dict, model_weights)
        # --------------------------------------------------------------------------------------------

        # ---- 6.1: Pinto el final
        end_date = datetime.datetime.now()
        self.CT.IT.info_print(f"fit_and_get_model_and_results terminado a las {end_date}. Ha tardado {end_date - start_date}", "light_magenta")

        # ---- 6.2: Almaceno el self.master_dict
        self.save_metrics_dict_json(self.master_result_dict, self.model_save_path, "complete_train_result")

        # ---- 6.3: Retorno la tupla
        return self.model, x_test_probs, x_train_probs, self.master_result_dict, model_weights

    def get_weights(self, model) -> List:
        """
        Implementacion del metodo que devuelve una lista de ndarrays con los pesos del modelo

        """

        return [model.feature_importances_] if hasattr(model, "feature_importances_") else [None]

    def get_model_save_path(self) -> str:
        return self.model_save_path

    def execute_shap_analysis(self, sample: bool = True, num_features_to_show: int = 100, num_sample: int = 500) -> pd.DataFrame:
        """
        Metodo para ejecutar el analisis SHAP
        :return:
        """

        return MlShapBinaryAndRegressorTools(self.x_test_df, self.model_name, self.model_save_path, self.model, sample, num_features_to_show, num_sample).run()

    # </editor-fold>

    # <editor-fold desc="Implementacion de propiedades abstractas   --------------------------------------------------------------------------------------------------------------">

    @property
    def verbose(self):
        verbose_dict: dict = {'show_all': 1, 'show_basics': 0, 'dont_show': -1}
        return verbose_dict[self._verbose]

    @property
    def metric(self):
        if self._metric not in ['mse', 'rmse', 'mae', 'msle', 'rmsle', 'mape']:
            raise ValueError(f'Invalid metric: {self._metric}')
        return self._metric

    @property
    def problem_type(self):
        problem_type_list: List[str] = ['regression', 'regression_l1', 'quantile', 'tweedie']
        if self._problem_type not in problem_type_list:
            raise ValueError(f'Invalid problem_type: {self._problem_type}. Must be in {problem_type_list}')
        return self._problem_type

    # </editor-fold>









from hx_ml_dl_tools import HxLightGbmRegressor, HxXtremeGradientBoostingRegressor, HxCatBoostRegressor
from hx_ml_dl_tools import HxDenseNeuralNetworkRegressor
from sklearn.datasets import load_diabetes, fetch_california_housing
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import StandardScaler
from hx_hyperparameters_builder import *
from info_tools import InfoTools
import pandas as pd


class DiabetesExample:
    def __init__(self):
        """
        Ejemplo de regresion de modelos de ML y DL para el ejemplo de la diabetes de sklearn
        """

        # --------------------------------------------------------------------------------------
        # -- 0: Instancio InfoTools y pinto la entrada
        # --------------------------------------------------------------------------------------

        # ---- 0.1: Instancia de InfoTools
        self.IT: InfoTools = InfoTools()

        # ---- 0.2: Pinto la entrada
        self.IT.header_print(f"DiabetesExample: Ejemplo de regresion con sklearn.datasets.load_diabetes utilizando hx_ml_dl_tools")

        # --------------------------------------------------------------------------------------
        # -- 1: Descargo el dataset y obtengo el data dict
        # --------------------------------------------------------------------------------------

        # ---- 1.1: Descargo el dataset
        self.diabetes_dataset = load_diabetes()

        # ---- 1.2: Transformo a dataframe (Lo requieren los modelos)
        self.x = pd.DataFrame(self.diabetes_dataset.data, columns=self.diabetes_dataset.feature_names)
        self.y = pd.DataFrame(self.diabetes_dataset.target, columns=["target"])

        # ---- 1.3: Divido en train y test
        self.x_train, self.x_test, self.y_train, self.y_test = train_test_split(self.x, self.y, test_size=0.2, random_state=42)

        # ---- 1.4: Creo el diccionario en el formato requerido por los modelos
        self.data_dict = {"x_train": self.x_train, "y_train": self.y_train, "x_test": self.x_test, "y_test": self.y_test}

        # --------------------------------------------------------------------------------------
        # -- 2: Creo a mano el diccionario de bins para estratificar
        # --------------------------------------------------------------------------------------

        # ---- 2.1: Este diccionario (OPCIONAL) sirve para categorizar las predicciones del regresor y poder observar donde acierta y falla mas
        self.bins: dict = {
            "bins": [0, 50, 100, 150, 200, 250, 300, 350, 400],
            "labels": ["0-50", "50-100", "100-150", "150-200", "200-250", "250-300", "300-350", "350-400"]
        }

        # ---- 2.2: Información del dataset
        self.IT.sub_intro_print(f"Dataset Diabetes cargado:")
        self.IT.info_print(f"Muestras totales: {self.x.shape[0]}")
        self.IT.info_print(f"Características: {self.x.shape[1]}")
        self.IT.info_print(f"Rango target: {self.y['target'].min():.2f} - {self.y['target'].max():.2f}")
        self.IT.info_print(f"Train samples: {self.x_train.shape[0]}")
        self.IT.info_print(f"Test samples: {self.x_test.shape[0]}")

        # ---- 2.3: Informacion de distribucion
        self.get_dataset_info()

    def execute_example(self):
        """
        Metodo que ejecuta en cadena los metodos de entrenamiento
        :return:
        """

        self._light_gbm_regressor()
        self._xg_boost_regressor()
        self._cat_boost_regressor()
        self._dnn_regressor()

    def get_dataset_info(self):
        """
        Metodo adicional para obtener información detallada del dataset
        """
        self.IT.sub_intro_print("Estadísticas del target:")

        self.IT.info_print(f"Media: {self.y['target'].mean():.2f}")
        self.IT.info_print(f"Mediana: {self.y['target'].median():.2f}")
        self.IT.info_print(f"Std: {self.y['target'].std():.2f}")
        self.IT.info_print(f"Min: {self.y['target'].min():.2f}")
        self.IT.info_print(f"Max: {self.y['target'].max():.2f}")

    # <editor-fold desc="Modelos   -----------------------------------------------------------------------------------------------------------------------------------------------">

    def _dnn_regressor(self):
        """
        Metodo para realizar el entrenamiento de la red neuronal densa de regresion
        :return:
        """

        # --------------------------------------------------------------------------------------
        # -- 1: Obtengo los hiperparametros que le voy a pasar
        # --------------------------------------------------------------------------------------

        # ---- 1.1: Opcion de pasarselos a mano
        # dnn_hyperparams: Dict[str, Union[int, float]] = {'units': 8, 'dropout': 0.5, 'learning_rate': 0.01, 'regularization': 0.1, 'epochs': 200}

        # ---- 1.2: Opcion de usar la clase HxDnnHyperparameterBuilder

        # ------ 1.2.1: Obtener parámetros aleatorios dentro del rango preestablecido en HxDnnHyperparameterBuilder
        # dnn_hyperparams: Dict[str, Union[int, float]] = HxDnnHyperparameterBuilder().get_random_hyperparams()

        # ------ 1.2.2: Pasarle los hiperparametros al constructor para que valide si hay alguno que se usa poco o no existe (USAMOS ESTA PARA EL EJEMPLO)
        dnn_hyperparams = HxDnnHyperparameterBuilder({'units': 8, 'dropout': 0.5, 'learning_rate': 0.01, 'regularization': 0.1, 'epochs': 200}).validate_and_get_hyperparams()

        # --------------------------------------------------------------------------------------
        # -- 2: Instancio la clase que me va a proporcionar un toolkit para ese modelo en concreto
        # --------------------------------------------------------------------------------------

        # ---- 2.1: Le paso los hiperparametros deseados (Entrar en la clase para ver detalle)
        dnn: HxDenseNeuralNetworkRegressor = HxDenseNeuralNetworkRegressor(
            self.data_dict,  # Diccionario con x_train, x_test, y_train, y_test
            dnn_hyperparams,  # Diccionario de hiperparametros
            'mae',  # Literal['mse', 'rmse', 'mae', 'msle', 'rmsle', 'mape'] (por defecto 'mse')
            'regression',  # Literal['regression'] = 'regression' (en este caso, solo puede ser regression)
            1,  # Literal[0, 1, 2, 3] = 1 ({0: "<", 1: ">", 2: "<>", 3: "><"}) Arquitectura de la red
            1,  # Capas ocultas de la red (Toda red tiene 1 capa de entrada + N capas ocultas + 1 capa de salida)
            "relu",  # Literal['relu', 'tanh', 'sigmoid'] = 'relu' (Funcion de activacion)
            "l1",  # Literal['l1', 'l2'] = 'l2'  (Tipo de regularizacion que se aplicará en caso de aplicar regularizacion)
            "adam",  # Literal['sgd', 'adam', 'rmsprop'] = 'adam'  (Optimizador que se va a utilizar)
            "mse",  # Literal['mse', 'mae', 'msle', 'mape', 'huber', 'log_cosh'] = 'mse' (Función de pérdida para el entrenamiento)
            True,  # bool (Incluir o no capas de dropout (en caso de no incluirse, el hiperparametro dropout no hace nada))
            True,  # bool (Incluir capa de batch normalization)
            15,  # Si en N epochs no ha mejorado la métrica, aplica un earlyStop callback y devuelve el mejor modelo hasta la fecha
            None,  # Path donde se almacenarán los resultados (por defecto None, ya que se crean internamente las estructuras de carpetas)
            self.bins,  #  dict | None = None (Diccionario de estratificacion)
            "dont_show",  # Literal['show_all', 'show_basics', 'dont_show'] = 'dont_show' (Cantidad de informacion adicional que se muestra en consola)
            42,  # Random state de incializacion de pesos para garantizar reproducibilidad
        )

        # ---- 2.2: [Metodo]: Permite entrenar el modelo, pintar las metricas en pantalla y devolver un diccionario de las metricas
        # dnn.fit_and_get_metrics(save_metrics_dict=False)

        # ---- 2.3: [Metodo]: Metodo principal de la clase que ejecuta internamente 'fit_and_get_metrics', almacena el modelo junto con graficos e info adicional
        dnn.fit_and_get_model_and_results()

        # ---- 2.4: [Metodo]: Permite ejecutar el analisis SHAP del modelo
        dnn.execute_shap_analysis(
            sample=True,  # Realizar un sampleo de la data antes de calcular SHAP (!! Es muy intensivo en cómputo si no se samplea !!)
            num_features_to_show=100,  # En los gráficos se van a mostrar las N características mas importantes (con mas peso)
            num_sample=200,  # Cantidad de muestra (filas) con las que se va a realizar el analisis
            background_sample=100,  # Número de instancias para background (explicador deep/grad).
        )

    def _light_gbm_regressor(self):
        """
        Metodo para realizar el entrenamiento del lgbm de regresion
        :return:
        """

        # --------------------------------------------------------------------------------------
        # -- 1: Obtengo los hiperparametros que le voy a pasar
        # --------------------------------------------------------------------------------------

        # ---- 1.1: En este caso, los obtengo aleatorios (en el ejemplo de la dnn están todas las opciones)
        lgbm_hyperparams: Dict[str, Union[int, float]] = HxLightgbmHyperparameterBuilder().get_random_hyperparams()

        # --------------------------------------------------------------------------------------
        # -- 2: Instancio la clase que me va a proporcionar un toolkit para ese modelo en concreto
        # --------------------------------------------------------------------------------------

        # ---- 2.1: Le paso los hiperparametros deseados (Entrar en la clase para ver detalle)
        lgbm: HxLightGbmRegressor = HxLightGbmRegressor(
            data_dict=self.data_dict,  # Diccionario con x_train, x_test, y_train, y_test
            hiperparams=lgbm_hyperparams,  # Diccionario con los hiperparametros del LGBM
            metric="mae",  # Literal['mse', 'rmse', 'mae', 'msle', 'rmsle', 'mape'], Metrica que vamos a evaluar
            problem_type="regression_l1",  # Literal['regression', 'regression_l1', 'quantile', 'tweedie'] = 'regression_l1' (funcion objetivo del lgbm)
            save_path=None,  # Path donde se almacenarán los resultados (por defecto None, ya que se crean internamente las estructuras de carpetas)
            bins=self.bins,  # dict | None = None (Diccionario de estratificacion)
            verbose="dont_show",  # Literal['show_all', 'show_basics', 'dont_show'] = 'dont_show' (Cantidad de informacion adicional que se muestra en consola)
            random_state=42,  # Random state de incializacion de pesos para garantizar reproducibilidad

        )

        # ---- 2.2: [Metodo]: Permite entrenar el modelo, pintar las metricas en pantalla y devolver un diccionario de las metricas
        # lgbm.fit_and_get_metrics(save_metrics_dict=False)

        # ---- 2.3: [Metodo]: Metodo principal de la clase que ejecuta internamente 'fit_and_get_metrics', almacena el modelo junto con graficos e info adicional
        lgbm.fit_and_get_model_and_results()

        # ---- 2.4: [Metodo]: Permite ejecutar el analisis SHAP del modelo
        lgbm.execute_shap_analysis(
            sample=True,  # Realizar un sampleo de la data antes de calcular SHAP (!! Es muy intensivo en cómputo si no se samplea !!)
            num_features_to_show=100,  # En los gráficos se van a mostrar las N características mas importantes (con mas peso)
            num_sample=200,  # Número de instancias para background (explicador deep/grad).
        )

    def _xg_boost_regressor(self):
        """
        Metodo para realizar el entrenamiento del xgboost de regresion
        :return:
        """

        # --------------------------------------------------------------------------------------
        # -- 1: Obtengo los hiperparametros que le voy a pasar
        # --------------------------------------------------------------------------------------

        # ---- 1.1: En este caso, los obtengo aleatorios (en el ejemplo de la dnn están todas las opciones)
        xgboost_hyperparams: Dict[str, Union[int, float]] = HxXgboostHyperparameterBuilder().get_random_hyperparams()

        # --------------------------------------------------------------------------------------
        # -- 2: Instancio la clase que me va a proporcionar un toolkit para ese modelo en concreto
        # --------------------------------------------------------------------------------------

        # ---- 2.1: Le paso los hiperparametros deseados (Entrar en la clase para ver detalle)
        xgboost: HxXtremeGradientBoostingRegressor = HxXtremeGradientBoostingRegressor(
            data_dict=self.data_dict,  # Diccionario con x_train, x_test, y_train, y_test
            hiperparams=xgboost_hyperparams,  # Diccionario con los hiperparametros del LGBM
            metric="mae",  # Literal['mse', 'rmse', 'mae', 'msle', 'rmsle', 'mape'], Metrica que vamos a evaluar
            problem_type="reg:absoluteerror",  # Literal['reg:quantileerror', 'reg:squarederror', 'reg:logistic', 'reg:squaredlogerror', 'reg:tweedie', 'reg:absoluteerror']
            save_path=None,  # Path donde se almacenarán los resultados (por defecto None, ya que se crean internamente las estructuras de carpetas)
            bins=self.bins,  # dict | None = None (Diccionario de estratificacion)
            verbose="dont_show",  # Literal['show_all', 'show_basics', 'dont_show'] = 'dont_show' (Cantidad de informacion adicional que se muestra en consola)
            random_state=42,  # Random state de incializacion de pesos para garantizar reproducibilidad

        )

        # ---- 2.2: [Metodo]: Permite entrenar el modelo, pintar las metricas en pantalla y devolver un diccionario de las metricas
        # xgboost.fit_and_get_metrics(save_metrics_dict=False)

        # ---- 2.3: [Metodo]: Metodo principal de la clase que ejecuta internamente 'fit_and_get_metrics', almacena el modelo junto con graficos e info adicional
        xgboost.fit_and_get_model_and_results()

        # ---- 2.4: [Metodo]: Permite ejecutar el analisis SHAP del modelo
        xgboost.execute_shap_analysis(
            sample=True,  # Realizar un sampleo de la data antes de calcular SHAP (!! Es muy intensivo en cómputo si no se samplea !!)
            num_features_to_show=100,  # En los gráficos se van a mostrar las N características mas importantes (con mas peso)
            num_sample=200,  # Número de instancias para background (explicador deep/grad).
        )

    def _cat_boost_regressor(self):
        """
        Metodo para realizar el entrenamiento del catboost de regresion
        :return:
        """

        # --------------------------------------------------------------------------------------
        # -- 1: Obtengo los hiperparametros que le voy a pasar
        # --------------------------------------------------------------------------------------

        # ---- 1.1: En este caso, los obtengo aleatorios (en el ejemplo de la dnn están todas las opciones)
        catboost_hyperparams: Dict[str, Union[int, float]] = HxCatBoostHyperparameterBuilder().get_random_hyperparams()

        # --------------------------------------------------------------------------------------
        # -- 2: Instancio la clase que me va a proporcionar un toolkit para ese modelo en concreto
        # --------------------------------------------------------------------------------------

        # ---- 2.1: Le paso los hiperparametros deseados (Entrar en la clase para ver detalle)
        catboost: HxCatBoostRegressor = HxCatBoostRegressor(
            data_dict=self.data_dict,  # Diccionario con x_train, x_test, y_train, y_test
            hiperparams=catboost_hyperparams,  # Diccionario con los hiperparametros del LGBM
            metric="mae",  # Literal['mse', 'rmse', 'mae', 'msle', 'rmsle', 'mape'], Metrica que vamos a evaluar
            problem_type="MAE",  # Literal['RMSE', 'MAE', 'Quantile', 'Tweedie'] = 'MAE'
            save_path=None,  # Path donde se almacenarán los resultados (por defecto None, ya que se crean internamente las estructuras de carpetas)
            bins=self.bins,  # dict | None = None (Diccionario de estratificacion)
            verbose="dont_show",  # Literal['show_all', 'show_basics', 'dont_show'] = 'dont_show' (Cantidad de informacion adicional que se muestra en consola)
            random_state=42,  # Random state de incializacion de pesos para garantizar reproducibilidad

        )

        # ---- 2.2: [Metodo]: Permite entrenar el modelo, pintar las metricas en pantalla y devolver un diccionario de las metricas
        # catboost.fit_and_get_metrics(save_metrics_dict=False)

        # ---- 2.3: [Metodo]: Metodo principal de la clase que ejecuta internamente 'fit_and_get_metrics', almacena el modelo junto con graficos e info adicional
        catboost.fit_and_get_model_and_results()

        # ---- 2.4: [Metodo]: Permite ejecutar el analisis SHAP del modelo
        catboost.execute_shap_analysis(
            sample=True,  # Realizar un sampleo de la data antes de calcular SHAP (!! Es muy intensivo en cómputo si no se samplea !!)
            num_features_to_show=100,  # En los gráficos se van a mostrar las N características mas importantes (con mas peso)
            num_sample=200,  # Número de instancias para background (explicador deep/grad).
        )

    # </editor-fold>


class CaliforniaHousingExample:
    def __init__(self):
        """
        Ejemplo de regresión de modelos de ML y DL para el dataset California Housing de sklearn
        """

        # --------------------------------------------------------------------------------------
        # -- 0: Instancio InfoTools y pinto la entrada
        # --------------------------------------------------------------------------------------

        # ---- 0.1: Instancia de InfoTools
        self.IT: InfoTools = InfoTools()

        # ---- 0.2: Pinto la entrada
        self.IT.header_print(f"CaliforniaHousingExample: Ejemplo de regresión con sklearn.datasets.fetch_california_housing utilizando hx_ml_dl_tools")

        # --------------------------------------------------------------------------------------
        # -- 1: Descargo el dataset y obtengo el data dict
        # --------------------------------------------------------------------------------------

        # ---- 1.1: Descargo el dataset
        self.california_dataset = fetch_california_housing()

        # ---- 1.2: Transformo a dataframe (Lo requieren los modelos)
        self.x = pd.DataFrame(self.california_dataset.data, columns=self.california_dataset.feature_names)
        self.y = pd.DataFrame(self.california_dataset.target, columns=["target"])

        # ---- 1.3: Preprocesamiento - Escalar características (IMPORTANTE para redes neuronales)
        scaler = StandardScaler()
        self.x_scaled = pd.DataFrame(scaler.fit_transform(self.x), columns=self.x.columns)

        # ---- 1.4: Divido en train y test
        self.x_train, self.x_test, self.y_train, self.y_test = train_test_split(self.x_scaled, self.y, test_size=0.2, random_state=42)

        # ---- 1.5: Creo el diccionario en el formato requerido por los modelos
        self.data_dict = {"x_train": self.x_train, "y_train": self.y_train, "x_test": self.x_test, "y_test": self.y_test}

        # --------------------------------------------------------------------------------------
        # -- 2: Creo a mano el diccionario de bins para estratificar (adaptado a California Housing)
        # --------------------------------------------------------------------------------------

        # ---- 2.1: Bins específicos para los precios de viviendas de California (en $100,000)
        self.bins: dict = {
            "bins": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
            "labels": ["0-1", "1-2", "2-3", "3-4", "4-5", "5-6", "6-7", "7-8", "8-9", "9-10"]
        }

        # ---- 2.2: Información del dataset
        self.IT.sub_intro_print(f"Dataset California Housing cargado:")
        self.IT.info_print(f"  - Muestras totales: {self.x.shape[0]}")
        self.IT.info_print(f"  - Características: {self.x.shape[1]}")
        self.IT.info_print(f"  - Rango target: {self.y['target'].min():.2f} - {self.y['target'].max():.2f} (en $100,000)")
        self.IT.info_print(f"  - Train samples: {self.x_train.shape[0]}")
        self.IT.info_print(f"  - Test samples: {self.x_test.shape[0]}")

        # ---- 2.3: Informacion de distribucion
        self.get_dataset_info()

    def execute_example(self):
        """
        Meodo que ejecuta en cadena los métodos de entrenamiento
        :return:
        """

        self.IT.sub_intro_print("Iniciando entrenamiento de modelos para California Housing")

        # Ejecutar modelos (puedes comentar los que no quieras ejecutar)
        self._light_gbm_regressor()
        self._xg_boost_regressor()
        self._cat_boost_regressor()
        self._dnn_regressor()

        self.IT.sub_intro_print("Entrenamiento completado para todos los modelos")

    def get_dataset_info(self):
        """
        Metodo adicional para obtener información detallada del dataset
        """
        self.IT.sub_intro_print("Estadísticas del target:")

        self.IT.info_print(f"Media: {self.y['target'].mean():.2f}")
        self.IT.info_print(f"Mediana: {self.y['target'].median():.2f}")
        self.IT.info_print(f"Std: {self.y['target'].std():.2f}")
        self.IT.info_print(f"Min: {self.y['target'].min():.2f}")
        self.IT.info_print(f"Max: {self.y['target'].max():.2f}")

    # <editor-fold desc="Modelos   -----------------------------------------------------------------------------------------------------------------------------------------------">

    def _dnn_regressor(self):
        """
        Metodo para realizar el entrenamiento de la red neuronal densa de regresión
        :return:
        """

        self.IT.sub_intro_print("Entrenando Red Neuronal Densa (DNN)")

        # --------------------------------------------------------------------------------------
        # -- 1: Obtengo los hiperparámetros optimizados para California Housing
        # --------------------------------------------------------------------------------------

        # ---- 1.1: Hiperparámetros específicos para este dataset más grande
        dnn_hyperparams = HxDnnHyperparameterBuilder({
            'units': 64,  # Más unidades para dataset más grande
            'dropout': 0.3,  # Dropout moderado
            'learning_rate': 0.001,  # Learning rate estándar
            'regularization': 0.01,  # Regularización ligera
            'epochs': 150  # Más épocas para convergencia
        }).validate_and_get_hyperparams()

        # --------------------------------------------------------------------------------------
        # -- 2: Instancio la clase DNN
        # --------------------------------------------------------------------------------------

        dnn: HxDenseNeuralNetworkRegressor = HxDenseNeuralNetworkRegressor(
            self.data_dict,  # Diccionario con los datos
            dnn_hyperparams,  # Hiperparámetros optimizados
            'mse',  # Métrica principal
            'regression',  # Tipo de problema
            2,  # Arquitectura "<>" (hourglass)
            3,  # 3 capas ocultas
            "relu",  # Función de activación
            "l2",  # Regularización L2
            "adam",  # Optimizador
            "mse",  # Función de pérdida
            True,  # Incluir dropout
            True,  # Incluir batch normalization
            20,  # Paciencia para early stopping
            None,  # Path de guardado (automático)
            self.bins,  # Bins para estratificación
            "dont_show",  # Verbose level
            42  # Random state
        )

        # ---- 2.2: Entrenar y obtener resultados completos
        dnn.fit_and_get_model_and_results()

        # ---- 2.3: Análisis SHAP (opcional - puede ser lento)
        dnn.execute_shap_analysis(sample=True, num_features_to_show=20, num_sample=50, background_sample=100)

    def _light_gbm_regressor(self):
        """
        Metodo para realizar el entrenamiento del LGBM de regresión
        :return:
        """

        self.IT.sub_intro_print("Entrenando LightGBM Regressor")

        # --------------------------------------------------------------------------------------
        # -- 1: Hiperparámetros para LightGBM
        # --------------------------------------------------------------------------------------

        lgbm_hyperparams: Dict[str, Union[int, float]] = HxLightgbmHyperparameterBuilder({
            'n_estimators': 200,
            'max_depth': 8,
            'learning_rate': 0.05,
            'subsample': 0.8,
            'colsample_bytree': 0.8
        }).validate_and_get_hyperparams()

        # --------------------------------------------------------------------------------------
        # -- 2: Instancio LightGBM
        # --------------------------------------------------------------------------------------

        lgbm: HxLightGbmRegressor = HxLightGbmRegressor(
            data_dict=self.data_dict,
            hiperparams=lgbm_hyperparams,
            metric="mse",
            problem_type="regression",
            save_path=None,
            bins=self.bins,
            verbose="show_basics",
            random_state=42
        )

        # ---- 2.2: Entrenar y obtener resultados
        lgbm.fit_and_get_model_and_results()

        # ---- 2.3: SHAP analysis
        lgbm.execute_shap_analysis(sample=True, num_features_to_show=20, num_sample=200)

    def _xg_boost_regressor(self):
        """
        Metodo para realizar el entrenamiento del XGBoost de regresión
        :return:
        """

        self.IT.sub_intro_print("Entrenando XGBoost Regressor")

        # --------------------------------------------------------------------------------------
        # -- 1: Hiperparámetros para XGBoost
        # --------------------------------------------------------------------------------------

        xgboost_hyperparams: Dict[str, Union[int, float]] = HxXgboostHyperparameterBuilder({
            'n_estimators': 200,
            'max_depth': 6,
            'learning_rate': 0.1,
            'subsample': 0.9,
            'colsample_bytree': 0.8
        }).validate_and_get_hyperparams()

        # --------------------------------------------------------------------------------------
        # -- 2: Instancio XGBoost
        # --------------------------------------------------------------------------------------

        xgboost: HxXtremeGradientBoostingRegressor = HxXtremeGradientBoostingRegressor(
            data_dict=self.data_dict,
            hiperparams=xgboost_hyperparams,
            metric="mse",
            problem_type="reg:squarederror",
            save_path=None,
            bins=self.bins,
            verbose="show_basics",
            random_state=42
        )

        # ---- 2.2: Entrenar y obtener resultados
        xgboost.fit_and_get_model_and_results()

        # ---- 2.3: SHAP analysis
        xgboost.execute_shap_analysis(sample=True, num_features_to_show=20, num_sample=200)

    def _cat_boost_regressor(self):
        """
        Metodo para realizar el entrenamiento del CatBoost de regresión
        :return:
        """

        self.IT.sub_intro_print("Entrenando CatBoost Regressor")

        # --------------------------------------------------------------------------------------
        # -- 1: Hiperparámetros para CatBoost
        # --------------------------------------------------------------------------------------

        catboost_hyperparams: Dict[str, Union[int, float]] = HxCatBoostHyperparameterBuilder({
            'iterations': 200,
            'depth': 6,
            'learning_rate': 0.05,
            'l2_leaf_reg': 3
        }).validate_and_get_hyperparams()

        # --------------------------------------------------------------------------------------
        # -- 2: Instancio CatBoost
        # --------------------------------------------------------------------------------------

        catboost: HxCatBoostRegressor = HxCatBoostRegressor(
            data_dict=self.data_dict,
            hiperparams=catboost_hyperparams,
            metric="mse",
            problem_type="RMSE",
            save_path=None,
            bins=self.bins,
            verbose="show_basics",
            random_state=42
        )

        # ---- 2.2: Entrenar y obtener resultados
        catboost.fit_and_get_model_and_results()

        # ---- 2.3: SHAP analysis
        catboost.execute_shap_analysis(sample=True, num_features_to_show=20, num_sample=200)

    # </editor-fold>



import numpy as np
from constants_and_tools import ConstantsAndTools
from typing import List, Literal, Any, Dict
from abc import ABC, abstractmethod
import pandas as pd


class HxPredictor(ABC):
    def __init__(self, df_to_predict: pd.DataFrame, model_directory: str, model_type: Literal['classifier', 'regressor'], verbose: bool = False):
        """
        Padre de las clases que se van a encargar de realizar predicciones tanto de machine learning como de deep learning.
        La idea es que a través del contenido de la carpeta que se crea con las instancias Hx de crear modelos, tengamos lo necesario
        para optimizar el proceso de predicción y tener transparenci en la información. Básicamente necesitamos 2 cosas del directorio:
        Por ejemplo, suponiendo un directorio 'LGBM_binary_classifier', en su contenido hay dos archivos que necesitamos:
        1: HxLightGbmClassifier.joblib --> Modelo
        2: complete_train_result.json --> Diccionario con columnas de predicción, métricas y resumen del entrenamiento
        :param df_to_predict: Dataframe imputado y escalado sobre el que tenemos que predecir
        :param model_directory: Carpeta donde está el modelo y el diccionario de entrenamiento
        :param model_type: Tipo de modelo (lo necesito para el predict)
        :param verbose: True para mostrar información del modelo y del proceso, False para que no muestre nada
        """

        self.CT: ConstantsAndTools = ConstantsAndTools()

        # -----------------------------------------------------------------------------------------
        # -- 1: Almaceno los parámetros en propiedades de clase y cargo modelo y diccionario
        # -----------------------------------------------------------------------------------------

        # ---- 1.1: Almaceno los parámetros
        self.df_to_predict: pd.DataFrame = df_to_predict.copy()
        self.model_directory: str = model_directory
        self.verbose: bool = verbose
        self.model_type: Literal['classifier', 'regressor'] = model_type

        # ---- 1.2: Obtengo el json del diccionario
        self._json_files_list: List[str] = self.validate_just_one_file(self.CT.OT.get_path_files_by_extension(self.model_directory, extension=".json"), ".json")

        # ---- 1.3: Almaceno el diccionario
        self._train_dict: Dict = [z for z in self.CT.PdT.load_json_in_dict(self._json_files_list[0]).values()][0]

        # -----------------------------------------------------------------------------------------
        # -- 2: Almaceno la informacion del diccionario en propiedades
        # -----------------------------------------------------------------------------------------

        # ---- 2.1: Columnas a predecir
        self._pred_cols: List[str] = self._train_dict["columns"]

        # ---- 2.2: Columna target
        self._target_col: str = self._train_dict["target"]

        # ---- 2.3: Hyperparams
        self._hyperparams: Dict[str, Any] = self._train_dict['best_params']

        # ---- 2.5: TODO: Metrics, modificar el diccionario para ademas de metrics, tener train metrics y test metrics
        self._test_metrics: Dict[str, Any] = self._train_dict['metrics']

        # -----------------------------------------------------------------------------------------
        # -- 3: Valido si las columnas de predicción están todas en el df que se ha pasado
        # -----------------------------------------------------------------------------------------

        # ---- 3.1: Ejecuto el metodo
        self.validate_prediction_cols_in_df()

    @staticmethod
    def validate_just_one_file(files_list: List[str], extension: str) -> List[str]:
        """
        Metodo que valida si en una lista hay un solo archivo y en caso contrario lanza excepcion
        :param extension:
        :param files_list:
        :return:
        """

        # -- COmpruebo el len de la lista, si es distinto de 1, lanzo excepciones
        if len(files_list) > 1:
            raise ValueError(f"HxPredictor Error: En el directorio que se ha facilitado, hay mas de un archivo del siguiente tipo: {extension}")
        elif len(files_list) < 1:
            raise ValueError(f"HxPredictor Error: En el directorio que se ha facilitado, no hay ficheros del siguiente tipo: {extension}")

        # -- Defino el caso correcto
        return files_list

    def validate_prediction_cols_in_df(self) -> None:
        """
        Este metodo va a validar si las columnas de prediccion existen
        :return:
        """

        # ---- Obtengo la lista de columnas con las que se entrenó el modelo y las comparo con las que trae el df
        pred_cols_not_in_df: List[str] = [z for z in self._pred_cols if z not in [x for x in self.df_to_predict.columns]]

        # ---- En caso de que alguna no esté, las pinto y lanzo excepcion
        if len(pred_cols_not_in_df) > 0:
            raise IndexError(f"Hay columnas con las que entrenó el modelo que no están en el df: {pred_cols_not_in_df}")

    def get_predictor_info(self) -> None:
        """
        Metodo que va a pintar las propiedades de interes del modelo
        :return:
        """

        self.CT.IT.sub_intro_print(f"Infomación del modelo")
        self.CT.IT.info_print(f"Target column: {self._target_col}")
        self.CT.IT.info_print(f"Hyperparameters: {self._hyperparams}")
        self.CT.IT.info_print(f"Test Metrics: {self._test_metrics}")

    @abstractmethod
    def predict(self) -> np.ndarray:
        pass